"""Text parser plugin for Atlassian Bitbucket log files.

This is for the atlassian-bitbucket.log file, one of multiple log files
produced by a Bitbucket DC/Server installation.

The default log format (logback pattern: %date %-5level [%thread] %request
%logger{36} %m%n%eThrowable) produces lines of the form:

  YYYY-MM-DD HH:MM:SS,mmm LEVEL [thread] logger_class message

With the optional %request context fields between [thread] and logger_class:

  YYYY-MM-DD HH:MM:SS,mmm LEVEL [thread] user request_identifier
  session_identifier ip_address "action" logger_class message

All %request fields are optional and may be absent. In practice, lines often
omit some or all of these fields depending on the context of the log event.

Also see:
  https://support.atlassian.com/bitbucket-data-center/kb/how-to-read-the-bitbucket-data-center-log-formats/
  https://support.atlassian.com/bitbucket-data-center/kb/how-to-change-the-bitbucket-application-log-format/
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class AtlassianBitbucketEventData(events.EventData):
  """Bitbucket application log event data.

  Attributes:
    body (str): the freeform body of the log line (the log message).
    ip_address (str): the client IP address associated with the request, if
        present in the log line.
    level (str): the logging level of the event, such as INFO, WARN or ERROR.
    logger_class (str): the abbreviated or full class name responsible for
        logging the event, such as c.a.b.m.r.DefaultRepositoryManager.
    request_action (str): the request action string, such as
        "TransactionService/Transact", if present.
    request_id (str): the unique request identifier, such as
        2CM38K4Fx339x113x2, if present.
    session_identifier (str): the session identifier, if present.
    thread (str): the JVM thread name from which the log event originated.
    user_name (str): the name of the user associated with the request, if
        present in the log line.
    written_time (dfdatetime.DateTimeValues): entry written date and time.
  """

  DATA_TYPE = 'atlassian:bitbucket:line'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.ip_address = None
    self.level = None
    self.logger_class = None
    self.request_action = None
    self.request_identifier = None
    self.session_identifier = None
    self.thread = None
    self.user_name = None
    self.written_time = None


class AtlassianBitbucketTextPlugin(interface.TextPlugin):
  """Text parser plugin for Atlassian Bitbucket log files."""

  NAME = 'atlassian_bitbucket'
  DATA_FORMAT = 'Atlassian Bitbucket log file'

  ENCODING = 'utf-8'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Bitbucket log levels.
  _BITBUCKET_LEVELS = ['DEBUG', 'ERROR', 'FATAL', 'INFO', 'TRACE', 'WARN']

  # Date and time format: 2020-09-08 07:53:45,084
  _DATE_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS +
      pyparsing.Suppress('-') + _TWO_DIGITS + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(',') + _THREE_DIGITS).set_results_name(
          'date_time')

  # Log level (DEBUG, ERROR, FATAL, INFO, TRACE, WARN).
  _LOG_LEVEL = pyparsing.oneOf(_BITBUCKET_LEVELS).set_results_name('level')

  # Thread name enclosed in brackets. Thread names do not contain ']'.
  _BITBUCKET_THREAD = (
      pyparsing.Suppress('[') +
      pyparsing.SkipTo(']').set_results_name('thread') +
      pyparsing.Suppress(']'))

  # Logger class, which can be abbreviated or full Java class name, such as:
  #     c.a.b.m.r.DefaultRepositoryManager
  #     com.atlassian.bitbucket.internal.boot.log.BuildInfoLogger
  #     o.h.i.ExceptionMapperStandardImpl
  #     org.hibernate.SQL
  #
  # The class name must start with a letter and contain at least one dot.
  _BITBUCKET_LOGGER = pyparsing.Regex(
      r'[a-zA-Z][a-zA-Z0-9_$]*(?:\.[a-zA-Z][a-zA-Z0-9_$]*)+').set_results_name(
          'logger_class')

  # Log message body: rest of line.
  _BITBUCKET_LOG_MESSAGE = pyparsing.SkipTo(
      pyparsing.LineEnd()).set_results_name('body')

  # The %request context block between [thread] and logger_class is optional.
  # It may contain: user, request_identifier, session_identifier, ip_address,
  # "action".
  # We capture everything between ']' and the logger class as original text and
  # parse it afterwards.
  _REQUEST_CONTEXT_TEXT = pyparsing.SkipTo(
      _BITBUCKET_LOGGER).set_results_name('request_context_text')

  # Guard that ensures the logger class is NOT preceded by '[', which
  # would indicate a Confluence-format log line where the logger is wrapped
  # in brackets: [thread] [logger.class] message.
  _NOT_BRACKET = pyparsing.NotAny(pyparsing.Literal('['))

  # Complete log line structure:
  # <timestamp> <level> [<thread>] [optional request context] <logger> <body>
  # _NOT_BRACKET before _REQUEST_CONTEXT_RAW ensures the first non-whitespace
  # character after [thread] is not '[', rejecting Confluence-format lines.
  _BITBUCKET_LOG_LINE = (
      _DATE_TIME + _LOG_LEVEL + _BITBUCKET_THREAD +
      _REQUEST_CONTEXT_TEXT + _BITBUCKET_LOGGER + _BITBUCKET_LOG_MESSAGE)

  _LINE_STRUCTURES = [('log_entry', _BITBUCKET_LOG_LINE)]

  # Sub-patterns for parsing the raw request context string.
  # Request identifier: alphanumeric token with 'x' separators.
  _REQUEST_IDENTIFIER = pyparsing.Regex(r'[0-9A-Za-z]{6,}x[0-9]+x[0-9]+x[0-9]+')

  # Session identifier: @/*/alphanumeric token, optionally comma-separated with
  # mesh.
  _SESSION_IDENTIFIER = pyparsing.Regex(
      r'[@*][A-Za-z0-9_x]+(?:,[A-Za-z0-9_x]+)?')

  # IPv4 or IPv6 address.
  _RE_IP_ADDRESS = pyparsing.Regex(
      r'(?:\d{1,3}\.){3}\d{1,3}|(?:[0-9a-fA-F]*:){2,7}[0-9a-fA-F]*')

  # Quoted action string.
  _RE_REQUEST_ACTION = pyparsing.QuotedString('"')

  # Username: word characters, dots, hyphens, slashes.
  _RE_USER_NAME = pyparsing.Regex(r'[A-Za-z][A-Za-z0-9._\-/]*')

  _REQUEST_CONTEXT_GRAMMAR = (
      pyparsing.Optional(_RE_USER_NAME.set_results_name('request_user')) +
      pyparsing.Optional(_REQUEST_IDENTIFIER.set_results_name(
          'request_identifier')) +
      pyparsing.Optional(_SESSION_IDENTIFIER.set_results_name(
          'session_identifier')) +
      pyparsing.Optional(_RE_IP_ADDRESS.set_results_name('ip_address')) +
      pyparsing.Optional(_RE_REQUEST_ACTION.set_results_name('request_action')))

  _LINE_STRUCTURES = [('log_entry', _BITBUCKET_LOG_LINE)]

  VERIFICATION_GRAMMAR = _BITBUCKET_LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key != 'log_entry':
      raise errors.ParseError(f'Unsupported structure: {key:s}')

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    request_context_text = self._GetValueFromStructure(
        structure, 'request_context_text', default_value='').strip()

    event_data = AtlassianBitbucketEventData()
    event_data.body = self._GetValueFromStructure(
        structure, 'body', default_value='').strip() or None
    event_data.level = self._GetValueFromStructure(structure, 'level')
    event_data.logger_class = self._GetValueFromStructure(
        structure, 'logger_class')
    event_data.thread = self._GetValueFromStructure(structure, 'thread')

    if request_context_text:
      request_context = self._REQUEST_CONTEXT_GRAMMAR.parse_string(
          request_context_text, parse_all=False)
      event_data.user_name = request_context.get('request_user') or None
      event_data.request_identifier = request_context.get(
          'request_identifier') or None
      event_data.session_identifier = request_context.get(
          'session_identifier') or None
      event_data.ip_address = request_context.get('ip_address') or None
      event_data.request_action = request_context.get('request_action') or None

    event_data.written_time = self._ParseTimeElements(time_elements_structure)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_structure)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    # Require at least 2 non-empty matching lines to guard against false
    # positives.

    verified_lines = 0
    for line in text_reader.lines.splitlines():
      if not line:
        continue

      try:
        structure = self._VerifyString(line)
      except errors.ParseError:
        return False

      verified_lines += 1
      if verified_lines >= 2:
        break

    if verified_lines < 2:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(AtlassianBitbucketTextPlugin)



"""Text parser plugins for TeamViewer log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class TeamViewerApplicationEventData(events.EventData):
  """TeamViewer application log event data.

  Attributes:
    body (str): body of the log entry.
    process_identifier (int): process identifier that generated the log entry.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
  """

  DATA_TYPE = 'teamviewer:application_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.body = None
    self.process_identifier = None
    self.recorded_time = None


class TeamViewerConnectionsIncomingEventData(events.EventData):
  """TeamViewer incoming connection log event data.

  Attributes:
    activity_type (str): Type of the activity, such as RemoteSupport or
        FileTransfer.
    connection_identifier (str): identifier of the connection, contains an UUID.
    display_name (string): The display name of the incoming connection source.
        Usually the computer name or the TeamViewer user name.
    end_time (dfdatetime.DateTimeValues): connection end time in UTC.
    local_account (str): The local user account associated with this activity.
    source_identifier (int): TeamViewer identifier of the incoming connection.
    start_time (dfdatetime.DateTimeValues): connection start time in UTC.
  """

  DATA_TYPE = 'teamviewer:connections_incoming:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.activity_type = None
    self.connection_identifier = None
    self.display_name = None
    self.end_time = None
    self.local_account = None
    self.source_identifier = None
    self.start_time = None


class TeamViewerConnectionsOutgoingEventData(events.EventData):
  """TeamViewer outgoing connection log event data.

  Attributes:
    activity_type (str): Type of the activity, such as RemoteSupport or
        FileTransfer.
    connection_identifier (str): identifier of the connection, contains a UUID.
    destination_identifier (int): TeamViewer identifier of the destination.
    end_time (dfdatetime.DateTimeValues): connection end time in UTC.
    local_account (str): The local user account associated
        with this activity.
    start_time (dfdatetime.DateTimeValues): connection start time in UTC.
  """

  DATA_TYPE = 'teamviewer:connections_outgoing:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.activity_type = None
    self.connection_identifier = None
    self.destination_identifier = None
    self.end_time = None
    self.local_account = None
    self.start_time = None


class TeamViewerApplicationLogTextPlugin(
    interface.TextPluginWithLineContinuation):
  """Text parser plugin for TeamViewer application log files."""

  NAME = 'teamviewer_application_log'
  DATA_FORMAT = 'TeamViewer application log file parser.'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TIMESTAMP = (
      _FOUR_DIGITS + pyparsing.Suppress('/') +
      _TWO_DIGITS + pyparsing.Suppress('/') +
      _TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress('.') +
      _THREE_DIGITS)

  _LOG_LINE = (
      _TIMESTAMP.set_results_name('timestamp') +
      pyparsing.Word(pyparsing.nums).set_parse_action(
          lambda tokens: int(tokens[0], 10)).set_results_name(
              'process_identifier') +
      # Field with unknown purpose.
      pyparsing.Suppress(pyparsing.Word(pyparsing.nums)) +
      # Field with unknown purpose.
      pyparsing.Suppress(pyparsing.Word(pyparsing.alphanums + '!')) +
      pyparsing.restOfLine().set_results_name('body') +
      pyparsing.Suppress(pyparsing.LineEnd()))

  VERIFICATION_GRAMMAR = _LOG_LINE

  _LINE_STRUCTURES = [('log_line', _LOG_LINE) ]

  def __init__(self):
    """Initializes a text parser plugin."""
    super().__init__()
    self._body_lines = None
    self._event_data = None

  def _ParseFinalize(self, parser_mediator):
    """Finalizes parsing.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    if self._event_data:
      self._event_data.body = ''.join(self._body_lines)
      self._body_lines = None

      parser_mediator.ProduceEventData(self._event_data)
      self._event_data = None

  def _ParseLogLine(self, structure):
    """Parses a log line.

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'timestamp')

    event_data = TeamViewerApplicationEventData()
    event_data.process_identifier = self._GetValueFromStructure(
         structure, 'process_identifier')
    event_data.recorded_time = self._ParseTimeElements(time_elements_structure)

    body = self._GetValueFromStructure(structure, 'body').strip()

    self._body_lines = [body]
    self._event_data = event_data

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == '_line_continuation':
      body = structure.strip()
      self._body_lines.append(body)

    else:
      if self._event_data:
        self._event_data.body = ''.join(self._body_lines)
        parser_mediator.ProduceEventData(self._event_data)

      self._ParseLogLine(structure)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      year, month, day, hours, minutes, seconds, milliseconds = (
          time_elements_structure)

      time_elements_tuple = (
          year, month, day, hours, minutes, seconds, milliseconds)

      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_tuple)
      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _ResetState(self):
    """Resets stored values."""
    self._body_lines = None
    self._event_data = None

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'timestamp')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    self._ResetState()

    return True


class TeamViewerConnectionsOutgoingLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for TeamViewer connections.txt log files."""

  NAME = 'teamviewer_connections_outgoing'
  DATA_FORMAT = 'TeamViewer connections.txt log file'

  _TEAMVIEWER_ID = pyparsing.Word(
      pyparsing.nums, min=8, max=11).set_parse_action(
          lambda tokens: int(tokens[0], 10))

  _NAME = pyparsing.Word(pyparsing.alphanums + '-' + '.' + '_')
  _GUID = pyparsing.Word(pyparsing.hexnums + '{' + '}' + '-')

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _DATE = (
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS +
      pyparsing.Suppress('-') + _FOUR_DIGITS)

  _TIME = (
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS)

  _DATE_TIME = _DATE + _TIME

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _TEAMVIEWER_ID.set_results_name('source_identifier') +
      _DATE_TIME.set_results_name('session_start_timestamp') +
      _DATE_TIME.set_results_name('session_end_timestamp') +
      _NAME.set_results_name('local_account') +
      pyparsing.Word(pyparsing.alphanums).set_results_name('activity_type') +
      _GUID.set_results_name('connection_identifier') +
      pyparsing.Suppress(pyparsing.LineEnd()))

  VERIFICATION_GRAMMAR = _LOG_LINE

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    start_time_elements = self._GetValueFromStructure(
        structure, 'session_start_timestamp')
    end_time_elements = self._GetValueFromStructure(
        structure, 'session_end_timestamp')

    event_data = TeamViewerConnectionsOutgoingEventData()
    event_data.activity_type = self._GetValueFromStructure(
        structure, 'activity_type')
    event_data.connection_identifier = self._GetValueFromStructure(
        structure, 'connection_identifier')
    event_data.end_time = self._ParseTimeElements(end_time_elements)
    event_data.local_account = self._GetValueFromStructure(
        structure, 'local_account')
    event_data.destination_identifier = self._GetValueFromStructure(
        structure, 'source_identifier')
    event_data.start_time = self._ParseTimeElements(start_time_elements)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      day_of_month, month, year, hours, minutes, seconds = (
          time_elements_structure)

      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple, time_zone_offset=0)
      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    start_time_elements = self._GetValueFromStructure(
        structure, 'session_start_timestamp')
    end_time_elements = self._GetValueFromStructure(
        structure, 'session_end_timestamp')

    try:
      self._ParseTimeElements(start_time_elements)
      self._ParseTimeElements(end_time_elements)
    except errors.ParseError:
      return False

    return True


class TeamViewerConnectionsIncomingLogTextPlugin(interface.TextPlugin):
  """Text parser plugin for TeamViewer connections_incoming.txt ."""

  NAME = 'teamviewer_connections_incoming'
  DATA_FORMAT = 'TeamViewer connections_incoming.txt log file'

  _TEAMVIEWER_ID = pyparsing.Word(
      pyparsing.nums, min=8, max=11).set_parse_action(
          lambda tokens: int(tokens[0], 10))

  _NAME = pyparsing.Word(pyparsing.alphanums + '-' + '.' + '_')
  _GUID = pyparsing.Word(pyparsing.hexnums + '{' + '}' + '-')

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _DATE = (
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS +
      pyparsing.Suppress('-') + _FOUR_DIGITS)

  _TIME = (
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS)

  _DATE_TIME = _DATE + _TIME

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE = (
      _TEAMVIEWER_ID.set_results_name('source_identifier') +
      _NAME.set_results_name('display_name') +
      _DATE_TIME.set_results_name('session_start_timestamp') +
      _DATE_TIME.set_results_name('session_end_timestamp') +
      _NAME.set_results_name('local_account') +
      pyparsing.Word(pyparsing.alphanums).set_results_name('activity_type') +
      _GUID.set_results_name('connection_identifier') +
      pyparsing.Suppress(pyparsing.LineEnd()))

  VERIFICATION_GRAMMAR = _LOG_LINE

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    start_time_elements = self._GetValueFromStructure(
        structure, 'session_start_timestamp')
    end_time_elements = self._GetValueFromStructure(
        structure, 'session_end_timestamp')

    event_data = TeamViewerConnectionsIncomingEventData()
    event_data.activity_type = self._GetValueFromStructure(
        structure, 'activity_type')
    event_data.connection_identifier = self._GetValueFromStructure(
        structure, 'connection_identifier')
    event_data.display_name = self._GetValueFromStructure(
        structure, 'display_name')
    event_data.end_time = self._ParseTimeElements(end_time_elements)
    event_data.local_account = self._GetValueFromStructure(
        structure, 'local_account')
    event_data.source_identifier = self._GetValueFromStructure(
        structure, 'source_identifier')
    event_data.start_time = self._ParseTimeElements(start_time_elements)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
           of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      day_of_month, month, year, hours, minutes, seconds = (
          time_elements_structure)

      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple, time_zone_offset=0)
      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    start_time_elements = self._GetValueFromStructure(
        structure, 'session_start_timestamp')
    end_time_elements = self._GetValueFromStructure(
        structure, 'session_end_timestamp')

    try:
      self._ParseTimeElements(start_time_elements)
      self._ParseTimeElements(end_time_elements)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugins([
    TeamViewerApplicationLogTextPlugin,
    TeamViewerConnectionsIncomingLogTextPlugin,
    TeamViewerConnectionsOutgoingLogTextPlugin])



"""Shared functionality for an extraction CLI tool."""

import datetime
import os
import time

import pytz

from dfvfs.analyzer import analyzer as dfvfs_analyzer
from dfvfs.lib import definitions as dfvfs_definitions
from dfvfs.lib import errors as dfvfs_errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context as dfvfs_context

# The following import makes sure the analyzers are registered.
from plaso import analyzers  # pylint: disable=unused-import

# The following import makes sure the parsers are registered.
from plaso import parsers  # pylint: disable=unused-import

from plaso.cli import logger
from plaso.cli import status_view
from plaso.cli import storage_media_tool
from plaso.cli import tool_options
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.containers import artifacts
from plaso.engine import configurations
from plaso.engine import engine
from plaso.filters import parser_filter
from plaso.helpers import language_tags
from plaso.lib import definitions
from plaso.lib import errors
from plaso.multi_process import extraction_engine as multi_extraction_engine
from plaso.parsers import manager as parsers_manager
from plaso.parsers import presets as parsers_presets
from plaso.single_process import extraction_engine as single_extraction_engine
from plaso.storage import factory as storage_factory


class ExtractionTool(
    storage_media_tool.StorageMediaTool,
    tool_options.HashersOptions,
    tool_options.ProfilingOptions,
    tool_options.StorageFileOptions):
  """Extraction CLI tool.

  Attributes:
    list_language_tags (bool): True if the language tags should be listed.
    list_time_zones (bool): True if the time zones should be listed.
  """

  _BYTES_IN_A_MIB = 1024 * 1024

  # Approximately 250 MB of queued items per worker.
  _DEFAULT_QUEUE_SIZE = 125000

  _PRESETS_FILE_NAME = 'presets.yaml'

  _SOURCE_TYPES_TO_PREPROCESS = frozenset([
      dfvfs_definitions.SOURCE_TYPE_DIRECTORY,
      dfvfs_definitions.SOURCE_TYPE_STORAGE_MEDIA_DEVICE,
      dfvfs_definitions.SOURCE_TYPE_STORAGE_MEDIA_IMAGE])

  _SUPPORTED_ARCHIVE_TYPES = {
      'iso9660': 'ISO-9660 disk image (.iso) file',
      'modi': 'MacOS disk image (.dmg) file',
      'tar': 'tape archive (.tar) file',
      'vhdi': 'Virtual Hard Disk image (.vhd, .vhdx) file',
      'zip': 'ZIP archive (.zip) file'}

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes an CLI tool.

    Args:
      input_reader (Optional[InputReader]): input reader, where None indicates
          that the stdin input reader should be used.
      output_writer (Optional[OutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    super().__init__(
        input_reader=input_reader, output_writer=output_writer)
    self._archive_types_string = 'none'
    self._artifacts_registry = None
    self._buffer_size = 0
    self._command_line_arguments = None
    self._enable_sigsegv_handler = False
    self._expanded_parser_filter_expression = None
    self._extract_winevt_resources = True
    self._extract_winreg_binary = True
    self._number_of_extraction_workers = 0
    self._parser_filter_expression = None
    self._preferred_codepage = None
    self._preferred_language = None
    self._preferred_time_zone = None
    self._preferred_year = None
    self._presets_file = None
    self._presets_manager = parsers_presets.ParserPresetsManager()
    self._process_compressed_streams = True
    self._process_memory_limit = None
    self._queue_size = self._DEFAULT_QUEUE_SIZE
    self._resolver_context = dfvfs_context.Context()
    self._single_process_mode = False
    self._status_view = status_view.StatusView(self._output_writer, self.NAME)
    self._status_view_file = 'status.info'
    self._status_view_interval = 0.5
    self._status_view_mode = status_view.StatusView.MODE_WINDOW
    self._storage_file_path = None
    self._storage_format = definitions.STORAGE_FORMAT_SQLITE
    self._task_storage_format = definitions.STORAGE_FORMAT_SQLITE
    self._temporary_directory = None
    self._worker_memory_limit = None
    self._worker_timeout = None
    self._yara_rules_string = None

    self.list_language_tags = False
    self.list_time_zones = False

  def _CheckStorageFile(self, storage_file_path, warn_about_existing=False):
    """Checks if the storage file path is valid.

    Args:
      storage_file_path (str): path of the storage file.
      warn_about_existing (bool): True if the user should be warned about
          the storage file already existing.

    Raises:
      BadConfigOption: if the storage file path is invalid.
    """
    if os.path.exists(storage_file_path):
      if not os.path.isfile(storage_file_path):
        raise errors.BadConfigOption((
            f'Storage file: {storage_file_path:s} already exists and is not '
            f'a file.'))

      if warn_about_existing:
        logger.warning('Appending to an already existing storage file.')

    dirname = os.path.dirname(storage_file_path)
    if not dirname:
      dirname = '.'

    # TODO: add a more thorough check to see if the storage file really is
    # a plaso storage file.

    if not os.access(dirname, os.W_OK):
      raise errors.BadConfigOption(
          f'Unable to write to storage file: {storage_file_path:s}')

  def _CreateExtractionEngine(self, single_process_mode):
    """Creates an extraction engine.

    Args:
      single_process_mode (bool): True if the engine should use single process
          mode.

    Returns:
      BaseEngine: extraction engine.
    """
    status_update_callback = (
        self._status_view.GetExtractionStatusUpdateCallback())

    if single_process_mode:
      extraction_engine = single_extraction_engine.SingleProcessEngine(
          status_update_callback=status_update_callback)
    else:
      extraction_engine = multi_extraction_engine.ExtractionMultiProcessEngine(
          number_of_worker_processes=self._number_of_extraction_workers,
          status_update_callback=status_update_callback,
          worker_memory_limit=self._worker_memory_limit,
          worker_timeout=self._worker_timeout)

    extraction_engine.SetStatusUpdateInterval(self._status_view_interval)

    return extraction_engine

  def _CreateExtractionProcessingConfiguration(self):
    """Creates an extraction processing configuration.

    Returns:
      ProcessingConfiguration: extraction processing configuration.
    """
    configuration = configurations.ProcessingConfiguration()
    configuration.artifact_definitions_path = self._artifact_definitions_path
    configuration.custom_artifacts_path = self._custom_artifacts_path
    configuration.data_location = self._data_location
    configuration.extraction.archive_types_string = self._archive_types_string
    configuration.artifact_filters = self._artifact_filters
    configuration.credentials = self._credential_configurations
    configuration.debug_output = self._debug_mode
    configuration.extraction.hasher_file_size_limit = (
        self._hasher_file_size_limit)
    configuration.extraction.extract_winevt_resources = (
        self._extract_winevt_resources)
    configuration.extraction.extract_winreg_binary = self._extract_winreg_binary
    configuration.extraction.hasher_names_string = self._hasher_names_string
    configuration.extraction.process_compressed_streams = (
        self._process_compressed_streams)
    configuration.extraction.yara_rules_string = self._yara_rules_string
    configuration.filter_file = self._filter_file
    configuration.log_filename = self._log_file
    configuration.parser_filter_expression = (
        self._expanded_parser_filter_expression)
    configuration.preferred_codepage = self._preferred_codepage
    configuration.preferred_language = self._preferred_language
    configuration.preferred_time_zone = self._preferred_time_zone
    configuration.preferred_year = self._preferred_year
    configuration.profiling.directory = self._profiling_directory
    configuration.profiling.sample_rate = self._profiling_sample_rate
    configuration.profiling.profilers = self._profilers
    configuration.task_storage_format = self._task_storage_format
    configuration.temporary_directory = self._temporary_directory

    return configuration

  def _GenerateStorageFileName(self):
    """Generates a name for the storage file.

    The result use a timestamp and the basename of the source path.

    Returns:
      str: a filename for the storage file in the form <time>-<source>.plaso

    Raises:
      BadConfigOption: raised if the source path is not set.
    """
    if not self._source_path:
      raise errors.BadConfigOption('Please define a source (--source).')

    timestamp = datetime.datetime.now()
    datetime_string = timestamp.strftime('%Y%m%dT%H%M%S')

    source_path = os.path.abspath(self._source_path)

    if source_path.endswith(os.path.sep):
      source_path = os.path.dirname(source_path)

    source_name = os.path.basename(source_path)

    if not source_name or source_name in ('/', '\\'):
      # The user passed the filesystem's root as source
      source_name = 'ROOT'

    return f'{datetime_string:s}-{source_name:s}.plaso'

  def _GetExpandedParserFilterExpression(self, system_configuration):
    """Determines the expanded parser filter expression.

    Args:
      system_configuration (SystemConfigurationArtifact): system configuration.

    Returns:
      str: expanded parser filter expression.

    Raises:
      BadConfigOption: if presets in the parser filter expression could not
          be expanded or if an invalid parser or plugin name is specified.
    """
    parser_filter_expression = self._parser_filter_expression
    if not parser_filter_expression and system_configuration:
      operating_system_artifact = artifacts.OperatingSystemArtifact(
          family=system_configuration.operating_system,
          product=system_configuration.operating_system_product,
          version=system_configuration.operating_system_version)

      preset_definitions = self._presets_manager.GetPresetsByOperatingSystem(
          operating_system_artifact)
      if preset_definitions:
        self._parser_filter_expression = ','.join([
            preset_definition.name
            for preset_definition in preset_definitions])

        logger.debug((
            f'Parser filter expression set to preset: '
            f'{self._parser_filter_expression:s}'))

    parser_filter_helper = parser_filter.ParserFilterExpressionHelper()

    try:
      parser_filter_expression = parser_filter_helper.ExpandPresets(
          self._presets_manager, self._parser_filter_expression)
      parser_filter_string = parser_filter_expression or 'N/A'
      logger.debug(f'Parser filter expression set to: {parser_filter_string:s}')
    except RuntimeError as exception:
      raise errors.BadConfigOption((
          f'Unable to expand presets in parser filter expression with '
          f'error: {exception!s}'))

    parser_elements, invalid_parser_elements = (
        parsers_manager.ParsersManager.CheckFilterExpression(
            parser_filter_expression))

    if invalid_parser_elements:
      invalid_parser_names_string = ','.join(invalid_parser_elements)
      raise errors.BadConfigOption((
          f'Unknown parser or plugin names in element(s): '
          f'"{invalid_parser_names_string:s}" of parser filter expression: '
          f'{parser_filter_expression:s}'))

    return ','.join(sorted(parser_elements))

  def _ParseExtractionOptions(self, options):
    """Parses the extraction options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['codepage', 'language'])

    # TODO: add preferred encoding

    self.list_language_tags = self._preferred_language == 'list'

    self._extract_winevt_resources = getattr(
        options, 'extract_winevt_resources', True)

    time_zone_string = self.ParseStringOption(options, 'timezone')
    if isinstance(time_zone_string, str):
      if time_zone_string.lower() == 'list':
        self.list_time_zones = True

      elif time_zone_string:
        try:
          pytz.timezone(time_zone_string)
        except pytz.UnknownTimeZoneError:
          raise errors.BadConfigOption(
              f'Unknown time zone: {time_zone_string:s}')

        self._preferred_time_zone = time_zone_string

  def _ParsePerformanceOptions(self, options):
    """Parses the performance options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    self._buffer_size = getattr(options, 'buffer_size', 0)
    if self._buffer_size:
      # TODO: turn this into a generic function that supports more size
      # suffixes both MB and MiB and also that does not allow m as a valid
      # indicator for MiB since m represents milli not Mega.
      try:
        if self._buffer_size[-1].lower() == 'm':
          self._buffer_size = int(self._buffer_size[:-1], 10)
          self._buffer_size *= self._BYTES_IN_A_MIB
        else:
          self._buffer_size = int(self._buffer_size, 10)
      except ValueError:
        raise errors.BadConfigOption(
            f'Invalid buffer size: {self._buffer_size!s}.')

    self._queue_size = self.ParseNumericOption(options, 'queue_size')

  def _ParseProcessingOptions(self, options):
    """Parses the processing options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    self._single_process_mode = getattr(options, 'single_process', False)

    argument_helper_names = [
        'process_resources', 'temporary_directory', 'vfs_backend', 'workers',
        'zeromq']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=argument_helper_names)

    if self._vfs_back_end == 'fsext':
      dfvfs_definitions.PREFERRED_EXT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_EXT)

    elif self._vfs_back_end == 'fsfat':
      dfvfs_definitions.PREFERRED_FAT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_FAT)

    elif self._vfs_back_end == 'fshfs':
      dfvfs_definitions.PREFERRED_HFS_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_HFS)

    elif self._vfs_back_end == 'fsntfs':
      dfvfs_definitions.PREFERRED_NTFS_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_NTFS)

    elif self._vfs_back_end == 'tsk':
      dfvfs_definitions.PREFERRED_EXT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_TSK)
      dfvfs_definitions.PREFERRED_FAT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_TSK)
      dfvfs_definitions.PREFERRED_GPT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_TSK_PARTITION)
      dfvfs_definitions.PREFERRED_HFS_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_TSK)
      dfvfs_definitions.PREFERRED_NTFS_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_TSK)

    elif self._vfs_back_end == 'vsgpt':
      dfvfs_definitions.PREFERRED_GPT_BACK_END = (
          dfvfs_definitions.TYPE_INDICATOR_GPT)

  def _ProcessSource(self, session, storage_writer):
    """Processes the source and extract events.

    Args:
      session (Session): session in which the source is processed.
      storage_writer (StorageWriter): storage writer to store extracted events.

    Returns:
      ProcessingStatus: processing status.

    Raises:
      BadConfigOption: if an invalid collection filter was specified.
    """
    single_process_mode = self._single_process_mode
    if self._source_type == dfvfs_definitions.SOURCE_TYPE_FILE:
      single_process_mode = True

    extraction_engine = self._CreateExtractionEngine(single_process_mode)

    extraction_engine.BuildArtifactsRegistry(
        self._artifact_definitions_path, self._custom_artifacts_path)

    source_configuration = artifacts.SourceConfigurationArtifact(
        path=self._source_path, source_type=self._source_type)

    # TODO: check if the source was processed previously.
    # TODO: add check for modification time of source.

    # If the source is a directory or a storage media image run pre-processing.

    system_configurations = []
    if self._source_type in self._SOURCE_TYPES_TO_PREPROCESS:
      try:
        logger.debug('Starting preprocessing.')

        system_configurations = extraction_engine.PreprocessSource(
            self._file_system_path_specs, storage_writer,
            resolver_context=self._resolver_context)

        logger.debug('Preprocessing done.')

      except OSError as exception:
        system_configurations = []

        logger.error(f'Unable to preprocess with error: {exception!s}')

      # TODO: check if the source was processed previously and if system
      # configuration differs.

    system_configuration = None
    if system_configurations:
      system_configuration = system_configurations[0]

    # TODO: add support for more than 1 system configuration.
    self._expanded_parser_filter_expression = (
        self._GetExpandedParserFilterExpression(system_configuration))

    enabled_parser_names = self._expanded_parser_filter_expression.split(',')

    number_of_enabled_parsers = len(enabled_parser_names)

    force_parser = False
    if (self._source_type == dfvfs_definitions.SOURCE_TYPE_FILE and
        number_of_enabled_parsers == 1):
      force_parser = True

      self._extract_winevt_resources = False

    elif ('winevt' not in enabled_parser_names and
          'winevtx' not in enabled_parser_names):
      self._extract_winevt_resources = False

    elif (self._extract_winevt_resources and
          'pe' not in enabled_parser_names):
      logger.warning(
          'A Windows EventLog parser is enabled in combination with '
          'extraction of Windows EventLog resources, but the Portable '
          'Executable (PE) parser is disabled. Therefore Windows EventLog '
          'resources cannot be extracted.')

      self._extract_winevt_resources = False

    processing_configuration = (
        self._CreateExtractionProcessingConfiguration())
    processing_configuration.force_parser = force_parser

    environment_variables = (
        extraction_engine.knowledge_base.GetEnvironmentVariables())
    user_accounts = list(storage_writer.GetAttributeContainers('user_account'))

    try:
      extraction_engine.BuildCollectionFilters(
          environment_variables, user_accounts,
          artifact_filter_names=self._artifact_filters,
          filter_file_path=self._filter_file)
    except errors.InvalidFilter as exception:
      raise errors.BadConfigOption(
          f'Unable to build collection filters with error: {exception!s}')

    session.artifact_filters = self._artifact_filters
    session.command_line_arguments = self._command_line_arguments
    session.debug_mode = self._debug_mode
    session.enabled_parser_names = enabled_parser_names
    session.extract_winevt_resources = self._extract_winevt_resources
    session.filter_file = self._filter_file
    session.parser_filter_expression = self._parser_filter_expression
    session.preferred_codepage = self._preferred_codepage
    session.preferred_encoding = self.preferred_encoding
    session.preferred_language = self._preferred_language or 'en-US'
    session.preferred_time_zone = self._preferred_time_zone
    session.preferred_year = self._preferred_year

    storage_writer.AddAttributeContainer(session)

    processing_status = None

    try:
      storage_writer.AddAttributeContainer(source_configuration)

      for system_configuration in system_configurations:
        storage_writer.AddAttributeContainer(system_configuration)

      if single_process_mode:
        logger.debug('Starting extraction in single process mode.')

        processing_status = extraction_engine.ProcessSource(
            storage_writer, self._resolver_context, processing_configuration,
            system_configurations, self._file_system_path_specs)

      else:
        logger.debug('Starting extraction in multi process mode.')

        # The method is named ProcessSourceMulti because pylint 2.6.0 and
        # later gets confused about keyword arguments when ProcessSource
        # is used.
        processing_status = extraction_engine.ProcessSourceMulti(
            storage_writer, session.identifier, processing_configuration,
            system_configurations, self._file_system_path_specs,
            enable_sigsegv_handler=self._enable_sigsegv_handler,
            storage_file_path=self._storage_file_path)

    finally:
      session.aborted = getattr(processing_status, 'aborted', True)
      session.completion_time = int(time.time() * 1000000)
      storage_writer.UpdateAttributeContainer(session)

    return processing_status

  def _ReadParserPresetsFromFile(self):
    """Reads the parser presets from the presets.yaml file.

    Raises:
      BadConfigOption: if the parser presets file cannot be read.
    """
    self._presets_file = os.path.join(
        self._data_location, self._PRESETS_FILE_NAME)
    if not os.path.isfile(self._presets_file):
      raise errors.BadConfigOption(
          f'No such parser presets file: {self._presets_file:s}')

    try:
      self._presets_manager.ReadFromFile(self._presets_file)
    except errors.MalformedPresetError as exception:
      raise errors.BadConfigOption(
          f'Unable to read parser presets from file with error: {exception!s}')

  def _ScanSourceForArchive(self, path_spec):
    """Determines if a path specification references an archive file.

    Args:
      path_spec (dfvfs.PathSpec): path specification of the data stream.

    Returns:
      dfvfs.PathSpec: path specification of the archive file or None if not
          an archive file.
    """
    try:
      type_indicators = (
          dfvfs_analyzer.Analyzer.GetCompressedStreamTypeIndicators(
              path_spec, resolver_context=self._resolver_context))
    except OSError:
      type_indicators = []

    if len(type_indicators) > 1:
      return False

    if type_indicators:
      type_indicator = type_indicators[0]
    else:
      type_indicator = None

    if type_indicator == dfvfs_definitions.TYPE_INDICATOR_BZIP2:
      path_spec = path_spec_factory.Factory.NewPathSpec(
          dfvfs_definitions.TYPE_INDICATOR_COMPRESSED_STREAM,
          compression_method=dfvfs_definitions.COMPRESSION_METHOD_BZIP2,
          parent=path_spec)

    elif type_indicator == dfvfs_definitions.TYPE_INDICATOR_GZIP:
      path_spec = path_spec_factory.Factory.NewPathSpec(
          dfvfs_definitions.TYPE_INDICATOR_GZIP, parent=path_spec)

    elif type_indicator == dfvfs_definitions.TYPE_INDICATOR_XZ:
      path_spec = path_spec_factory.Factory.NewPathSpec(
          dfvfs_definitions.TYPE_INDICATOR_COMPRESSED_STREAM,
          compression_method=dfvfs_definitions.COMPRESSION_METHOD_XZ,
          parent=path_spec)

    try:
      type_indicators = dfvfs_analyzer.Analyzer.GetArchiveTypeIndicators(
          path_spec, resolver_context=self._resolver_context)
    except OSError:
      return None

    if len(type_indicators) != 1:
      return None

    return path_spec_factory.Factory.NewPathSpec(
        type_indicators[0], location='/', parent=path_spec)

  def AddExtractionOptions(self, argument_group):
    """Adds the extraction options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        argument_group, names=['codepage', 'language'])

    # Note defaults here are None so we can determine if an option was set.

    argument_group.add_argument(
        '--no_extract_winevt_resources', '--no-extract-winevt-resources',
        dest='extract_winevt_resources', action='store_false', default=True,
        help=('Do not extract Windows EventLog resources such as event '
              'message template strings. By default Windows EventLog '
              'resources will be extracted when a Windows EventLog parser '
              'is enabled.'))

    # TODO: add preferred encoding

    argument_group.add_argument(
        '-z', '--zone', '--timezone', dest='timezone', action='store',
        metavar='TIME_ZONE', type=str, default=None, help=(
            'preferred time zone of extracted date and time values that are '
            'stored without a time zone indicator. The time zone is determined '
            'based on the source data where possible otherwise it will default '
            'to UTC. Use "list" to see a list of available time zones.'))

  def AddPerformanceOptions(self, argument_group):
    """Adds the performance options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    argument_group.add_argument(
        '--buffer_size', '--buffer-size', '--bs', dest='buffer_size',
        action='store', default=0, help=(
            'The buffer size for the output (defaults to 196MiB).'))

    argument_group.add_argument(
        '--queue_size', '--queue-size', dest='queue_size', action='store',
        default=0, help=(
            f'The maximum number of queued items per worker (defaults to '
            f'{self._DEFAULT_QUEUE_SIZE:d})'))

  def AddProcessingOptions(self, argument_group):
    """Adds the processing options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    argument_group.add_argument(
        '--single_process', '--single-process', dest='single_process',
        action='store_true', default=False, help=(
            'Indicate that the tool should run in a single process.'))

    argument_helper_names = [
        'temporary_directory', 'vfs_backend', 'workers', 'zeromq']
    if self._CanEnforceProcessMemoryLimit():
      argument_helper_names.append('process_resources')
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        argument_group, names=argument_helper_names)

  def ExtractEventsFromSources(self):
    """Processes the sources and extracts events.

    Raises:
      BadConfigOption: if the storage file path is invalid, or the storage
          format not supported, or there was a failure to writing to the
          storage.
      OSError: if the extraction engine could not write to the storage.
      SourceScannerError: if the source scanner could not find a supported
          file system.
      UserAbort: if the user initiated an abort.
    """
    self._CheckStorageFile(self._storage_file_path, warn_about_existing=True)

    try:
      self.ScanSource(self._source_path)
    except dfvfs_errors.UserAbort as exception:
      raise errors.UserAbort(exception)

    if self._source_type == dfvfs_definitions.SOURCE_TYPE_FILE:
      archive_path_spec = self._ScanSourceForArchive(
          self._file_system_path_specs[0])
      if archive_path_spec:
        self._file_system_path_specs = [archive_path_spec]
        self._source_type = definitions.SOURCE_TYPE_ARCHIVE

    self._status_view.SetMode(self._status_view_mode)
    self._status_view.SetStatusFile(self._status_view_file)
    self._status_view.SetSourceInformation(
        self._source_path, self._source_type,
        artifact_filters=self._artifact_filters,
        filter_file=self._filter_file)

    self._output_writer.Write('\n')
    self._status_view.PrintExtractionStatusHeader(None)
    self._output_writer.Write('Processing started.\n')

    # TODO: attach processing configuration to session?
    session = engine.BaseEngine.CreateSession()

    storage_writer = storage_factory.StorageFactory.CreateStorageWriter(
        self._storage_format)
    if not storage_writer:
      raise errors.BadConfigOption(
          f'Unsupported storage format: {self._storage_format:s}')

    try:
      storage_writer.Open(path=self._storage_file_path)
    except OSError as exception:
      raise OSError(f'Unable to open storage with error: {exception!s}')

    processing_status = None
    number_of_extraction_warnings = 0

    try:
      stored_number_of_extraction_warnings = (
          storage_writer.GetNumberOfAttributeContainers('extraction_warning'))

      try:
        processing_status = self._ProcessSource(session, storage_writer)

      finally:
        number_of_extraction_warnings = (
            storage_writer.GetNumberOfAttributeContainers(
                'extraction_warning') - stored_number_of_extraction_warnings)

    except OSError as exception:
      raise OSError(f'Unable to write to storage with error: {exception!s}')

    finally:
      storage_writer.Close()

    self._status_view.PrintExtractionSummary(
        processing_status, number_of_extraction_warnings)

  def ListArchiveTypes(self):
    """Lists information about supported archive types."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Archive and storage media image types')

    for name, description in sorted(self._SUPPORTED_ARCHIVE_TYPES.items()):
      table_view.AddRow([name, description])

    table_view.Write(self._output_writer)

  def ListLanguageTags(self):
    """Lists the language tags."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Language tag', 'Description'],
        title='Language tags')
    for language_tag, description in (
        language_tags.LanguageTagHelper.GetLanguages()):
      table_view.AddRow([language_tag, description])
    table_view.Write(self._output_writer)

  def ListParsersAndPlugins(self):
    """Lists information about the available parsers and plugins."""
    parsers_information = parsers_manager.ParsersManager.GetParsersInformation()

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Parsers')

    for name, description in sorted(parsers_information):
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)

    parser_names = parsers_manager.ParsersManager.GetNamesOfParsersWithPlugins()
    for parser_name in parser_names:
      plugins_information = (
          parsers_manager.ParsersManager.GetParserPluginsInformation(
              parser_filter_expression=parser_name))

      table_title = f'Parser plugins: {parser_name:s}'
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, column_names=['Name', 'Description'],
          title=table_title)
      for name, description in sorted(plugins_information):
        table_view.AddRow([name, description])
      table_view.Write(self._output_writer)

    title = 'Parser presets'
    if self._presets_file:
      source_path = os.path.dirname(os.path.dirname(os.path.dirname(
          os.path.abspath(__file__))))

      presets_file = self._presets_file
      if presets_file.startswith(source_path):
        presets_file = presets_file[len(source_path) + 1:]

      title = f'{title:s} ({presets_file:s})'

    presets_information = self._presets_manager.GetPresetsInformation()

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Parsers and plugins'],
        title=title)
    for name, description in sorted(presets_information):
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)



"""Plist parser plugin for Mac OS startup item plist files."""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSStartupItemEventData(events.EventData):
  """Mac OS startup item event data.

  Attributes:
    description (str): description of the startup item.
    order_preference (str): startup order preference.
    provides (list[str]): names of services provided by the startup item.
    requires (list[str]): services required prior to this startup item.
    uses (list[str]): services that should be started before this startup item.
  """

  DATA_TYPE = 'macos:startup_item:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.description = None
    self.order_preference = None
    self.provides = None
    self.requires = None
    self.uses = None


class MacOSStartupItemPlugin(interface.PlistPlugin):
  """Plist parser plugin for Mac OS startup item plist files."""

  NAME = 'macos_startup_item_plist'
  DATA_FORMAT = 'Mac OS startup item plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('StartupParameters.plist')])

  PLIST_KEYS = frozenset([])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extracts startup item information from the plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    event_data = MacOSStartupItemEventData()
    event_data.description = top_level.get('Description')
    event_data.order_preference = top_level.get('OrderPreference')
    event_data.provides = top_level.get('Provides')
    event_data.requires = top_level.get('Requires')
    event_data.uses = top_level.get('Uses')
    parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSStartupItemPlugin)



"""Text parser plugin for iOS sysdiagnose logd files (logd.0.log)."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class IOSSysdiagnoseLogdData(events.EventData):
  """iOS sysdiagnose logd event data.

  Attributes:
    body (str): body of the event line.
    logger (str): name of the process that generated the event.
    written_time (dfdatetime.DateTimeValues): date and time the log entry was
        written.
  """

  DATA_TYPE = 'ios:sysdiagnose:logd:line'

  def __init__(self):
    """Initializes iOS sysdiagnose logd event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.logger = None
    self.written_time = None


class IOSSysdiagnoseLogdTextPlugin(interface.TextPlugin):
  """Text parser plugin for iOS sysdiagnose logd files (logd.0.log)."""

  NAME = 'ios_logd'
  DATA_FORMAT = 'iOS sysdiagnose logd file'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TIME_ZONE_OFFSET = (
      pyparsing.Word('+-', exact=1) + _TWO_DIGITS + _TWO_DIGITS)

  # Date and time values are formatted as: 2021-08-11 05:50:23-0700
  _DATE_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      _TIME_ZONE_OFFSET)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOGGER = pyparsing.Combine(
      pyparsing.Word(pyparsing.alphas + '_') + pyparsing.Literal('[') +
      pyparsing.Word(pyparsing.nums) + pyparsing.Literal(']'))

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      _LOGGER.set_results_name('logger') + pyparsing.Suppress(': ') +
      pyparsing.restOfLine().set_results_name('body') + _END_OF_LINE)

  _LINE_STRUCTURES = [('log_entry', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    event_data = IOSSysdiagnoseLogdData()
    event_data.body = self._GetValueFromStructure(
        structure, 'body', default_value='')
    event_data.logger = self._GetValueFromStructure(
        structure, 'logger', default_value='')
    event_data.written_time = self._ParseTimeElements(time_elements_structure)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      (year, month, day_of_month, hours, minutes, seconds, time_zone_sign,
       time_zone_hours, time_zone_minutes) = time_elements_structure

      time_zone_offset = (time_zone_hours * 60) + time_zone_minutes
      if time_zone_sign == '-':
        time_zone_offset *= -1

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      return dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple,
          time_zone_offset=time_zone_offset)

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(IOSSysdiagnoseLogdTextPlugin)



"""Text parser plugin for Microsoft IIS log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class IISEventData(events.EventData):
  """IIS log event data.

  Attributes:
    cs_cookie (str): Content of a sent or received cookie.
    cs_host (str): HTTP host header name.
    cs_referrer (str): Site that referred to the requested site.
    cs_uri_query (str): URI query that was requested.
    cs_username (str): Username of the authenticated user that accessed
        the server, where anonymous users are indicated by a hyphen.
    dest_ip (str): IP address of the server that generated the logged activity.
    dest_port (str): Server port number.
    http_method (str): HTTP request method, such as GET or POST.
    http_status (str): HTTP status code that was returned by the server.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    protocol_version (str): HTTP protocol version that was used.
    received_bytes (str): Number of bytes received and processed by the server.
    requested_uri_stem (str): File requested, such as index.php or Default.htm
    s_computername (str): Name of the server that generated the logged activity.
    sc_substatus (str):  HTTP substatus error code that was returned by the
        server.
    sc_win32_status (str): Windows status code of the server.
    sent_bytes (str): Number of bytes sent by the server.
    source_ip (str): IP address of the client that made the request.
    s_sitename (str): Service name and instance number that was running on the
        client.
    time_taken (str): Time taken, in milliseconds, to process the request.
    user_agent (str): User agent that was used.
  """

  DATA_TYPE = 'iis:log:line'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cs_cookie = None
    self.cs_host = None
    self.cs_referrer = None
    self.cs_uri_query = None
    self.cs_username = None
    self.dest_ip = None
    self.dest_port = None
    self.http_method = None
    self.http_status = None
    self.last_written_time = None
    self.protocol_version = None
    self.received_bytes = None
    self.requested_uri_stem = None
    self.s_computername = None
    self.sc_substatus = None
    self.sc_win32_status = None
    self.sent_bytes = None
    self.source_ip = None
    self.s_sitename = None
    self.time_taken = None
    self.user_agent = None


class WinIISTextPlugin(interface.TextPlugin):
  """Text parser plugin for Microsoft IIS log files."""

  NAME = 'winiis'
  DATA_FORMAT = 'Microsoft IIS log file'

  # Log file are all extended ASCII encoded unless UTF-8 is explicitly enabled.
  # TODO: fix
  ENCODING = 'utf-8'

  _BLANK = pyparsing.Literal('-')

  _HTTP_METHOD = pyparsing.Word(pyparsing.alphanums + '-_') | _BLANK

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10)) | _BLANK

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _IP_ADDRESS = (
      pyparsing.pyparsing_common.ipv4_address |
      pyparsing.pyparsing_common.ipv6_address | _BLANK)

  PORT = pyparsing.Word(pyparsing.nums, max=6).set_parse_action(
      lambda tokens: int(tokens[0], 10)) | _BLANK

  # Username can consist of: "domain.username", "domain\username",
  # "domain\user$", "domain/user", "user@domain" or "-" for an anonymous user.
  _USERNAME = pyparsing.Word(pyparsing.alphanums + '-.\\$@/') | _BLANK

  _URI_SAFE_CHARACTERS = '/.?&+;_=()-:,%'

  _URI = pyparsing.Word(pyparsing.alphanums + _URI_SAFE_CHARACTERS) | _BLANK

  _URI_STEM = (pyparsing.Word(
      pyparsing.alphanums + _URI_SAFE_CHARACTERS + '$') | _BLANK)

  _UA = pyparsing.Word(
      pyparsing.alphanums + _URI_SAFE_CHARACTERS + '[]') | _BLANK

  _COOKIE = pyparsing.Word(
      pyparsing.alphanums + _URI_SAFE_CHARACTERS + '@{}"\\') | _BLANK

  # Per https://blogs.iis.net/nazim/use-of-special-characters-like-in-an-iis-url
  # IIS does not require that a query comply with RFC1738 restrictions on valid
  # URI characters
  _QUERY = (pyparsing.Word(
      pyparsing.alphanums + _URI_SAFE_CHARACTERS + '{}|\\^~[]`\'"<>@$') |
           _BLANK)

  _DATE = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS)

  _TIME = (
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS)

  _DATE_TIME_METADATA = (
      pyparsing.Suppress('Date: ') +
      _DATE.set_results_name('date') +
      _TIME.set_results_name('time'))

  _FIELDS_METADATA = (
      pyparsing.Suppress('Fields: ') +
      pyparsing.restOfLine().set_results_name('fields'))

  _METADATA = (
      _DATE_TIME_METADATA | _FIELDS_METADATA | pyparsing.restOfLine())

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _COMMENT_LOG_LINE = pyparsing.Suppress('#') + _METADATA + _END_OF_LINE

  # IIS 6.x fields: date time s-sitename s-ip cs-method cs-uri-stem
  # cs-uri-query s-port cs-username c-ip cs(User-Agent) sc-status
  # sc-substatus sc-win32-status

  _IIS_6_0_LOG_LINE = (
      _DATE.set_results_name('date') +
      _TIME.set_results_name('time') +
      _URI.set_results_name('s_sitename') +
      _IP_ADDRESS.set_results_name('dest_ip') +
      _HTTP_METHOD.set_results_name('http_method') +
      _URI.set_results_name('cs_uri_stem') +
      _URI.set_results_name('cs_uri_query') +
      PORT.set_results_name('dest_port') +
      _USERNAME.set_results_name('cs_username') +
      _IP_ADDRESS.set_results_name('source_ip') +
      _UA.set_results_name('user_agent') +
      _INTEGER.set_results_name('sc_status') +
      _INTEGER.set_results_name('sc_substatus') +
      _INTEGER.set_results_name('sc_win32_status') +
      _END_OF_LINE)

  # IIS 7.x fields: date time s-ip cs-method cs-uri-stem cs-uri-query
  # s-port cs-username c-ip cs(User-Agent) sc-status sc-substatus
  # sc-win32-status time-taken

  _LOG_LINE_STRUCTURES = {}

  # Common fields. Set results name with underscores, not hyphens because regex
  # will not pick them up.

  _LOG_LINE_STRUCTURES['date'] = _DATE.set_results_name('date')
  _LOG_LINE_STRUCTURES['time'] = _TIME.set_results_name('time')
  _LOG_LINE_STRUCTURES['s-sitename'] = _URI.set_results_name('s_sitename')
  _LOG_LINE_STRUCTURES['s-ip'] = _IP_ADDRESS.set_results_name('dest_ip')
  _LOG_LINE_STRUCTURES['cs-method'] = _HTTP_METHOD.set_results_name(
      'http_method')
  _LOG_LINE_STRUCTURES['cs-uri-stem'] = _URI_STEM.set_results_name(
      'requested_uri_stem')
  _LOG_LINE_STRUCTURES['cs-uri-query'] = _QUERY.set_results_name('cs_uri_query')
  _LOG_LINE_STRUCTURES['s-port'] = PORT.set_results_name('dest_port')
  _LOG_LINE_STRUCTURES['cs-username'] = _USERNAME.set_results_name(
      'cs_username')
  _LOG_LINE_STRUCTURES['c-ip'] = _IP_ADDRESS.set_results_name('source_ip')
  _LOG_LINE_STRUCTURES['cs(User-Agent)'] = _UA.set_results_name('user_agent')
  _LOG_LINE_STRUCTURES['sc-status'] = _INTEGER.set_results_name('http_status')
  _LOG_LINE_STRUCTURES['sc-substatus'] = _INTEGER.set_results_name(
      'sc_substatus')
  _LOG_LINE_STRUCTURES['sc-win32-status'] = _INTEGER.set_results_name(
      'sc_win32_status')

  # Less common fields.

  _LOG_LINE_STRUCTURES['s-computername'] = _URI.set_results_name(
      's_computername')
  _LOG_LINE_STRUCTURES['sc-bytes'] = _INTEGER.set_results_name('sent_bytes')
  _LOG_LINE_STRUCTURES['cs-bytes'] = _INTEGER.set_results_name('received_bytes')
  _LOG_LINE_STRUCTURES['time-taken'] = _INTEGER.set_results_name('time_taken')
  _LOG_LINE_STRUCTURES['cs-version'] = _URI.set_results_name('protocol_version')
  _LOG_LINE_STRUCTURES['cs-host'] = _URI.set_results_name('cs_host')
  _LOG_LINE_STRUCTURES['cs(Cookie)'] = _COOKIE.set_results_name('cs_cookie')
  _LOG_LINE_STRUCTURES['cs(Referrer)'] = _URI.set_results_name('cs_referrer')
  _LOG_LINE_STRUCTURES['cs(Referer)'] = _URI.set_results_name('cs_referrer')

  # Define the available log line structures. Default to the IIS v. 6.0
  # common format.

  _HEADER_GRAMMAR = pyparsing.OneOrMore(_COMMENT_LOG_LINE)

  _LINE_STRUCTURES = [('log_line', _IIS_6_0_LOG_LINE)]

  _COMMENT_SOFTWARE_LINE = (
      pyparsing.Regex(
          '#Software: Microsoft Internet Information Services [0-9]+.[0-9]+') +
      _END_OF_LINE)

  VERIFICATION_GRAMMAR = (
      pyparsing.ZeroOrMore(
          pyparsing.Regex('#(Date|Fields|Version): .*') + _END_OF_LINE) +
      _COMMENT_SOFTWARE_LINE)

  VERIFICATION_LITERALS = [
      '#Software: Microsoft Internet Information Services ']

  def __init__(self):
    """Initializes a parser."""
    super().__init__()
    self._day_of_month = None
    self._month = None
    self._year = None

  def _ParseFieldsMetadata(self, parser_mediator, fields):
    """Parses the fields metadata and updates the log line definition to match.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      fields (str): field definitions.
    """
    log_line_structure = pyparsing.Empty()
    for member in fields.split(' '):
      if not member:
        continue

      field_structure = self._LOG_LINE_STRUCTURES.get(member)
      if not field_structure:
        field_structure = self._URI
        parser_mediator.ProduceExtractionWarning(
            f'missing definition for field: {member:s} defaulting to URI')

      log_line_structure += field_structure

    log_line_structure += self._END_OF_LINE

    self._SetLineStructures([('log_line', log_line_structure)])

  def _ParseHeader(self, parser_mediator, text_reader):
    """Parses a text-log file header.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Raises:
      ParseError: when the header cannot be parsed.
    """
    try:
      structure_generator = self._HEADER_GRAMMAR.scan_string(
          text_reader.lines, max_matches=1)
      structure, start, end = next(structure_generator)

    except StopIteration:
      structure = None

    except pyparsing.ParseException as exception:
      raise errors.ParseError(exception)

    if not structure or start != 0:
      raise errors.ParseError('No match found.')

    date_elements_tuple = self._GetValueFromStructure(structure, 'date')
    if date_elements_tuple:
      self._year, self._month, self._day_of_month = date_elements_tuple

    fields = self._GetValueFromStructure(structure, 'fields', default_value='')
    fields = fields.strip()
    if fields:
      self._ParseFieldsMetadata(parser_mediator, fields)

    text_reader.SkipAhead(end)

  def _ParseLogLine(self, parser_mediator, structure):
    """Parse a single log line.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): tokens from a parsed log line.
    """
    event_data = IISEventData()
    event_data.cs_cookie = self._GetValueFromStructure(structure, 'cs_cookie')
    event_data.cs_host = self._GetValueFromStructure(structure, 'cs_host')
    event_data.cs_referrer = self._GetValueFromStructure(
        structure, 'cs_referrer')
    event_data.cs_uri_query = self._GetValueFromStructure(
        structure, 'cs_uri_query')
    event_data.cs_username = self._GetValueFromStructure(
        structure, 'cs_username')
    event_data.dest_ip = self._GetValueFromStructure(structure, 'dest_ip')
    event_data.dest_port = self._GetValueFromStructure(structure, 'dest_port')
    event_data.http_method = self._GetValueFromStructure(
        structure, 'http_method')
    event_data.http_status = self._GetValueFromStructure(
        structure, 'http_status')
    event_data.protocol_version = self._GetValueFromStructure(
        structure, 'protocol_version')
    event_data.last_written_time = self._ParseTimeElements(structure)
    event_data.received_bytes = self._GetValueFromStructure(
        structure, 'received_bytes')
    event_data.requested_uri_stem = self._GetValueFromStructure(
        structure, 'requested_uri_stem')
    event_data.s_computername = self._GetValueFromStructure(
        structure, 's_computername')
    event_data.sc_substatus = self._GetValueFromStructure(
        structure, 'sc_substatus')
    event_data.sc_win32_status = self._GetValueFromStructure(
        structure, 'sc_win32_status')
    event_data.sent_bytes = self._GetValueFromStructure(structure, 'sent_bytes')
    event_data.source_ip = self._GetValueFromStructure(structure, 'source_ip')
    event_data.s_sitename = self._GetValueFromStructure(structure, 's_sitename')
    event_data.time_taken = self._GetValueFromStructure(structure, 'time_taken')
    event_data.user_agent = self._GetValueFromStructure(structure, 'user_agent')

    parser_mediator.ProduceEventData(event_data)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    self._ParseLogLine(parser_mediator, structure)

  def _ParseTimeElements(self, structure):
    """Parses date and time elements of a log line.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      time_elements_structure = self._GetValueFromStructure(structure, 'time')

      hours, minutes, seconds = time_elements_structure

      date_elements_structure = self._GetValueFromStructure(structure, 'date')
      if date_elements_structure:
        year, month, day_of_month = date_elements_structure

        time_elements_tuple = (
            year, month, day_of_month, hours, minutes, seconds)
      else:
        time_elements_tuple = (
             self._year, self._month, self._day_of_month, hours, minutes,
             seconds)

      return dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _ResetState(self):
    """Resets stored values."""
    self._day_of_month = None
    self._month = None
    self._year = None

    self._SetLineStructures(self._LINE_STRUCTURES)

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    self._ResetState()

    return True


text_parser.TextLogParser.RegisterPlugin(WinIISTextPlugin)



"""Fake (in-memory only) storage writer for testing."""

from plaso.lib import definitions
from plaso.storage import writer
from plaso.storage.fake import fake_store


class FakeStorageWriter(writer.StorageWriter):
  """Fake (in-memory only) storage writer object.

  Attributes:
    task_completion (TaskCompletion): task completion attribute container.
    task_start (TaskStart): task start attribute container.
  """

  def __init__(self, storage_type=definitions.STORAGE_TYPE_SESSION):
    """Initializes a storage writer object.

    Args:
      storage_type (Optional[str]): storage type.
    """
    super().__init__(storage_type=storage_type)
    self._first_written_event_data_index = 0
    self._first_written_event_source_index = 0
    self._written_event_data_index = 0
    self._written_event_source_index = 0
    self.task_completion = None
    self.task_start = None

  def GetFirstWrittenEventData(self):
    """Retrieves the first event data that was written after open.

    Using GetFirstWrittenEventData and GetNextWrittenEventData newly
    added event data can be retrieved in order of addition.

    Returns:
      EventData: event data or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_data = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_DATA, self._first_written_event_data_index)
    self._written_event_data_index = self._first_written_event_data_index + 1
    return event_data

  def GetNextWrittenEventData(self):
    """Retrieves the next event data that was written after open.

    Returns:
      EventData: event data or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_data = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_DATA, self._written_event_data_index)
    self._written_event_data_index += 1
    return event_data

  def GetFirstWrittenEventSource(self):
    """Retrieves the first event source that was written after open.

    Using GetFirstWrittenEventSource and GetNextWrittenEventSource newly
    added event sources can be retrieved in order of addition.

    Returns:
      EventSource: event source or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_source = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_SOURCE,
        self._first_written_event_source_index)
    self._written_event_source_index = (
        self._first_written_event_source_index + 1)
    return event_source

  def GetNextWrittenEventSource(self):
    """Retrieves the next event source that was written after open.

    Returns:
      EventSource: event source or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_source = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_SOURCE, self._written_event_source_index)
    self._written_event_source_index += 1
    return event_source

  def Open(self, **unused_kwargs):
    """Opens the storage writer.

    Raises:
      OSError: if the storage writer is already opened.
    """
    if self._store:
      raise OSError('Storage writer already opened.')

    self._store = fake_store.FakeStore()
    self._store.Open()

    self._first_written_event_data_index = 0
    self._first_written_event_source_index = 0
    self._written_event_data_index = 0
    self._written_event_source_index = 0



"""The preprocess mediator."""

import pytz

from plaso.containers import warnings
from plaso.helpers.windows import eventlog_providers
from plaso.helpers.windows import time_zones
from plaso.preprocessors import logger


class PreprocessMediator:
  """Preprocess mediator.

  Attributes:
    code_page (str): code page.
    hostname (HostnameArtifact): hostname.
    language (str): language.
    time_zone (datetime.tzinfo): time zone.
  """

  def __init__(self, storage_writer):
    """Initializes a preprocess mediator.

    Args:
      storage_writer (StorageWriter): storage writer, to store preprocessing
          information in.
    """
    super().__init__()
    self._available_time_zones = {}
    self._environment_variables = {}
    self._file_entry = None
    self._storage_writer = storage_writer
    self._windows_eventlog_providers_helper = (
        eventlog_providers.WindowsEventLogProvidersHelper())
    self._windows_eventlog_providers = {}
    self._windows_eventlog_providers_by_identifier = {}
    self._values = {}

    self.code_page = None
    self.hostname = None
    self.language = None
    self.time_zone = None

  def AddArtifact(self, artifact_attribute_container):
    """Adds a pre-processing artifact attribute container.

    Args:
      artifact_attribute_container (ArtifactAttributeContainer): artifact
          attribute container.
    """
    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(artifact_attribute_container)

  def AddEnvironmentVariable(self, environment_variable_artifact):
    """Adds an environment variable.

    Args:
      environment_variable_artifact (EnvironmentVariableArtifact): environment
          variable artifact.

    Raises:
      KeyError: if the environment variable already exists.
    """
    logger.debug((
        f'setting environment variable: {environment_variable_artifact.name:s} '
        f'to: "{environment_variable_artifact.value:s}"'))

    name = environment_variable_artifact.name.upper()
    if name in self._environment_variables:
      raise KeyError((
          f'Environment variable: {environment_variable_artifact.name:s} '
          f'already exists.'))

    self._environment_variables[name] = environment_variable_artifact

    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(environment_variable_artifact)

  def AddHostname(self, hostname_artifact):
    """Adds a hostname.

    Args:
      hostname_artifact (HostnameArtifact): hostname artifact.
    """
    # TODO: change storage and pre-processor to handle more than 1 hostname.
    self.hostname = hostname_artifact

  def AddTimeZoneInformation(self, time_zone_artifact):
    """Adds a time zone defined by the operating system.

    Args:
      time_zone_artifact (TimeZoneArtifact): time zone artifact.

    Raises:
      KeyError: if the time zone already exists.
    """
    if time_zone_artifact.name in self._available_time_zones:
      raise KeyError(f'Time zone: {time_zone_artifact.name:s} already exists.')

    self._available_time_zones[time_zone_artifact.name] = time_zone_artifact

    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(time_zone_artifact)

  def AddUserAccount(self, user_account):
    """Adds an user account.

    Args:
      user_account (UserAccountArtifact): user account artifact.

    Raises:
      KeyError: if the user account already exists.
    """
    if not user_account.username:
      logger.debug(f'adding user account: {user_account.identifier:s}')
    else:
      logger.debug(
          f'adding user account: {user_account.username:s} '
          f'({user_account.identifier:s})')

    if self._storage_writer:
      self._storage_writer.AddAttributeContainer(user_account)

  def AddWindowsEventLogProvider(self, windows_eventlog_provider):
    """Adds a Windows EventLog provider.

    Args:
      windows_eventlog_provider (WindowsEventLogProviderArtifact): Windows
          EventLog provider.

    Raises:
      KeyError: if the Windows EventLog provider already exists.
    """
    existing_provider = None
    provider_identifier = windows_eventlog_provider.identifier
    log_source = windows_eventlog_provider.log_sources[0]

    if provider_identifier:
      existing_provider = self._windows_eventlog_providers_by_identifier.get(
          provider_identifier, None)

    if not existing_provider:
      existing_provider = self._windows_eventlog_providers.get(log_source)

    if existing_provider:
      self._windows_eventlog_providers_helper.Merge(
          existing_provider, windows_eventlog_provider)

      if self._storage_writer:
        self._storage_writer.UpdateAttributeContainer(existing_provider)

    else:
      self._windows_eventlog_providers_helper.NormalizeMessageFiles(
          windows_eventlog_provider)

      self._windows_eventlog_providers[log_source] = windows_eventlog_provider

      if self._storage_writer:
        self._storage_writer.AddAttributeContainer(windows_eventlog_provider)

      if provider_identifier:
        self._windows_eventlog_providers_by_identifier[provider_identifier] = (
            windows_eventlog_provider)

  def GetEnvironmentVariable(self, name):
    """Retrieves an environment variable.

    Args:
      name (str): name of the environment variable.

    Returns:
      EnvironmentVariableArtifact: environment variable artifact or None
          if there was no value set for the given name.
    """
    name = name.upper()
    return self._environment_variables.get(name)

  def GetEnvironmentVariables(self):
    """Retrieves the environment variables.

    Returns:
      list[EnvironmentVariableArtifact]: environment variable artifacts.
    """
    return self._environment_variables.values()

  def GetValue(self, identifier):
    """Retrieves a value by identifier.

    Args:
      identifier (str): case insensitive unique identifier for the value.

    Returns:
      object: value or None if not available.
    """
    identifier = identifier.lower()
    return self._values.get(identifier)

  def GetValues(self):
    """Retrieves the values.

    Returns:
      list[tuple[str, object]]: values.
    """
    return self._values.items()

  def ProducePreprocessingWarning(self, plugin_name, message):
    """Produces a preprocessing warning.

    Args:
      plugin_name (str): name of the preprocess plugin.
      message (str): message of the warning.
    """
    if self._storage_writer:
      path_spec = None
      if self._file_entry:
        path_spec = self._file_entry.path_spec

      warning = warnings.PreprocessingWarning(
          message=message, path_spec=path_spec, plugin_name=plugin_name)
      self._storage_writer.AddAttributeContainer(warning)

    logger.debug(f'[{plugin_name:s}] {message:s}')

  def Reset(self):
    """Resets the values stored in the mediator."""
    self._available_time_zones = {}
    self._environment_variables = {}
    self._file_entry = None
    self._windows_eventlog_providers = {}
    self._windows_eventlog_providers_by_identifier = {}
    self._values = {}

    self.code_page = None
    self.hostname = None
    self.language = None
    self.time_zone = None

  def SetCodePage(self, code_page):
    """Sets the code page.

    Args:
      code_page (str): code_page.

    Raises:
      ValueError: if the code page is not supported.
    """
    logger.debug(f'setting code page to: "{code_page:s}"')
    self.code_page = code_page

  def SetFileEntry(self, file_entry):
    """Sets the active file entry.

    Args:
      file_entry (dfvfs.FileEntry): file entry.
    """
    self._file_entry = file_entry

  def SetLanguage(self, language):
    """Sets the language.

    Args:
      language (str): language.

    Raises:
      ValueError: if the language is not supported.
    """
    self.language = language

  def SetTimeZone(self, time_zone):
    """Sets the time zone.

    Args:
      time_zone (str): time zone.

    Raises:
      ValueError: if the time zone is not supported.
    """
    # Get the "normalized" name of a Windows time zone name.
    if time_zone.startswith('@tzres.dll,'):
      mui_form_time_zones = {
          time_zone_artifact.mui_form: time_zone_artifact.name
          for time_zone_artifact in self._available_time_zones.values()}

      time_zone = mui_form_time_zones.get(time_zone, time_zone)
    else:
      localized_time_zones = {
          time_zone_artifact.localized_name: time_zone_artifact.name
          for time_zone_artifact in self._available_time_zones.values()}

      time_zone = localized_time_zones.get(time_zone, time_zone)

    # Map a Windows time zone name to a Python time zone name.
    time_zone = time_zones.WINDOWS_TIME_ZONES.get(time_zone, time_zone)

    try:
      self.time_zone = pytz.timezone(time_zone)
    except pytz.UnknownTimeZoneError:
      raise ValueError(f'Unsupported time zone: {time_zone!s}')

  def SetValue(self, identifier, value):
    """Sets a value by identifier.

    Args:
      identifier (str): case insensitive unique identifier for the value.
      value (object): value.

    Raises:
      TypeError: if the identifier is not a string type.
    """
    identifier = identifier.lower()
    if identifier not in self._values:
      self._values[identifier] = value



"""Text parser plugin for Atlassian Confluence log files.

This is for the atlassian-confluence.log file, one of multiple log files
produced by a Confluence DC/Server installation.
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class AtlassianConfluenceEventData(events.EventData):
  """Confluence event data.

  Attributes:
    body (str): the freeform body of the log line.
    level (str): the logging level of the event.
    logger_class (str): the Confluence class responsible for logging.
    logger_method (str): name of the method within the class.
    thread (str): the Confluence thread from which the log event originated.
    written_time (dfdatetime.DateTimeValues): entry written date and time.
  """

  DATA_TYPE = 'atlassian:confluence:line'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.level = None
    self.logger_class = None
    self.logger_method = None
    self.thread = None
    self.written_time = None


class AtlassianConfluenceTextPlugin(interface.TextPlugin):
  """Text parser plugin for Atlassian Confluence log files."""

  NAME = 'atlassian_confluence'
  DATA_FORMAT = 'Atlassian Confluence log file'

  ENCODING = 'utf-8'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Confluence log levels
  _CONFLUENCE_LEVELS = ['DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL']

  # Date and time format: 2022-07-12 01:08:59,489
  # Format: YYYY-MM-DD HH:MM:SS,mmm (comma-separated milliseconds)
  _DATE_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS +
      pyparsing.Suppress('-') + _TWO_DIGITS + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(',') + _THREE_DIGITS).set_results_name(
          'date_time')

  # Log level (DEBUG, INFO, WARN, ERROR, FATAL)
  _LOG_LEVEL = pyparsing.oneOf(_CONFLUENCE_LEVELS).set_results_name('level')

  # Thread name enclosed in brackets: [Catalina-utility-1]
  # or [hz.confluence.event-3]
  # Allows alphanumerics, hyphens, underscores, dots, colons, and spaces
  _CONFLUENCE_THREAD = (
      pyparsing.Suppress('[') +
      pyparsing.Word(
          pyparsing.alphanums + '-_.:' + ' ').set_results_name('thread') +
      pyparsing.Suppress(']'))

  # Logger class name enclosed in brackets
  # [confluence.cluster.hazelcast.HazelcastClusterManager]
  _CONFLUENCE_LOGGER = (
      pyparsing.Suppress('[') +
      pyparsing.SkipTo(']').set_results_name('logger_class') +
      pyparsing.Suppress(']'))

  # Logger method name: startCluster, memberAdded, <init>,
  # lambda$getCompletedStatuses$0
  _CONFLUENCE_LOGGER_METHOD = (
      pyparsing.Word(
          pyparsing.alphanums + '_$<>').set_results_name('logger_method'))

  # Log message body - rest of line
  _CONFLUENCE_LOG_MESSAGE = pyparsing.SkipTo(
      pyparsing.LineEnd()).set_results_name('body')

  # Complete log line structure
  # Format: <timestamp> <level> [<thread>] [<logger_class>] <method> <message>
  _CONFLUENCE_LOG_LINE = (
      _DATE_TIME + _LOG_LEVEL + _CONFLUENCE_THREAD + _CONFLUENCE_LOGGER +
      _CONFLUENCE_LOGGER_METHOD + _CONFLUENCE_LOG_MESSAGE)

  _LINE_STRUCTURES = [('log_entry', _CONFLUENCE_LOG_LINE)]

  VERIFICATION_GRAMMAR = _CONFLUENCE_LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key != 'log_entry':
      raise errors.ParseError(
          f'Unable to parse record, unknown structure: {key:s}')

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    event_data = AtlassianConfluenceEventData()
    event_data.body = self._GetValueFromStructure(
        structure, 'body', default_value='').strip() or None
    event_data.level = self._GetValueFromStructure(structure, 'level')
    event_data.logger_class = self._GetValueFromStructure(
        structure, 'logger_class')
    event_data.logger_method = self._GetValueFromStructure(
        structure, 'logger_method')
    event_data.thread = self._GetValueFromStructure(structure, 'thread')
    event_data.written_time = self._ParseTimeElements(time_elements_structure)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_structure)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(AtlassianConfluenceTextPlugin)



"""JSON-L parser plugin for Docker container configuration files."""

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import jsonl_parser
from plaso.parsers.jsonl_plugins import interface


class DockerContainerConfigurationEventData(events.EventData):
  """Docker container configuration event data.

  Attributes:
    action (str): whether the container was created, started, or finished.
    container_identifier (str): identifier of the container (SHA256).
    container_name (str): name of the container.
    creation_time (dfdatetime.DateTimeValues): date and time the container
        was created (added).
    end_time (dfdatetime.DateTimeValues): date and time the container
        was stopped.
    start_time (dfdatetime.DateTimeValues): date and time the container
        was started.
  """

  DATA_TYPE = 'docker:container:configuration'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.action = None
    self.container_identifier = None
    self.container_name = None
    self.creation_time = None
    self.end_time = None
    self.start_time = None


class DockerContainerConfigurationJSONLPlugin(interface.JSONLPlugin):
  """JSON-L parser plugin for Docker container configuration files.

  This parser handles per Docker container configuration files stored in:
  DOCKER_DIR/containers/<container_identifier>/config.json
  """

  NAME = 'docker_container_config'
  DATA_FORMAT = 'Docker container configuration file'

  def _ParseISO8601DateTimeString(self, parser_mediator, json_dict, name):
    """Parses an ISO8601 date and time string.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary.
      name (str): name of the value to retrieve.

    Returns:
      dfdatetime.TimeElementsInMicroseconds: date and time value or None if
          not available.
    """
    iso8601_string = self._GetJSONValue(json_dict, name)
    if not iso8601_string:
      return None

    # A FinishedAt date and time value of 0001-01-01T00:00:00Z represents
    # that the container is still running.
    if name == 'FinishedAt' and iso8601_string == '0001-01-01T00:00:00Z':
      return None

    try:
      date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()
      date_time.CopyFromStringISO8601(iso8601_string)
    except ValueError as exception:
      parser_mediator.ProduceExtractionWarning((
          f'Unable to parse value: {name:s} ISO8601 string: '
          f'{iso8601_string:s} with error: {exception!s}'))
      return None

    return date_time

  def _ParseRecord(self, parser_mediator, json_dict):
    """Parses a Docker container configuration record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary of the configuration record.
    """
    json_state = self._GetJSONValue(json_dict, 'State', default_value={})
    configuration = self._GetJSONValue(json_dict, 'Config', default_value={})

    event_data = DockerContainerConfigurationEventData()
    event_data.container_identifier = self._GetJSONValue(json_dict, 'ID')
    event_data.container_name = self._GetJSONValue(
        configuration, 'Hostname', default_value='Unknown container name')
    event_data.creation_time = self._ParseISO8601DateTimeString(
        parser_mediator, json_dict, 'Created')
    event_data.end_time = self._ParseISO8601DateTimeString(
        parser_mediator, json_state, 'FinishedAt')
    event_data.start_time = self._ParseISO8601DateTimeString(
        parser_mediator, json_state, 'StartedAt')

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, json_dict):
    """Check if the record has the minimal structure required by the plugin.

    Args:
      json_dict (dict): JSON dictionary of the configuration record.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    configuration = self._GetJSONValue(json_dict, 'Config')
    driver = self._GetJSONValue(json_dict, 'Driver')
    identifier = self._GetJSONValue(json_dict, 'ID')

    if None in (configuration, driver, identifier):
      return False

    # TODO: validate format of container identifier.

    return True


jsonl_parser.JSONLParser.RegisterPlugin(DockerContainerConfigurationJSONLPlugin)



"""The processing engine."""

import os

from artifacts import errors as artifacts_errors
from artifacts import reader as artifacts_reader
from artifacts import registry as artifacts_registry

from dfvfs.lib import errors as dfvfs_errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import resolver as path_spec_resolver

from plaso.containers import artifacts
from plaso.containers import sessions
from plaso.engine import artifact_filters
from plaso.engine import knowledge_base
from plaso.engine import logger
from plaso.engine import path_filters
from plaso.engine import processing_status
from plaso.engine import profilers
from plaso.engine import yaml_filter_file
from plaso.lib import errors
from plaso.preprocessors import manager as preprocess_manager
from plaso.preprocessors import mediator as preprocess_mediator


class BaseEngine:
  """Processing engine interface.

  Attributes:
    knowledge_base (KnowledgeBase): knowledge base.
  """

  _WINDOWS_REGISTRY_FILES_ARTIFACT_NAMES = [
      'WindowsSystemRegistryFiles', 'WindowsUserRegistryFiles']

  def __init__(self):
    """Initializes an engine."""
    super().__init__()
    self._abort = False
    self._analyzers_profiler = None
    self._artifacts_registry = None
    self._artifacts_trie = None
    self._excluded_file_system_find_specs = None
    self._included_file_system_find_specs = None
    self._memory_profiler = None
    self._name = 'Main'
    self._processing_status = processing_status.ProcessingStatus()
    self._processing_profiler = None
    self._registry_find_specs = None
    self._serializers_profiler = None
    # The interval of status updates in number of seconds.
    self._status_update_interval = 0.5
    self._storage_profiler = None
    self._task_queue_profiler = None

    self.knowledge_base = knowledge_base.KnowledgeBase()

  def _StartProfiling(self, configuration):
    """Starts profiling.

    Args:
      configuration (ProfilingConfiguration): profiling configuration.
    """
    if not configuration:
      return

    if configuration.HaveProfileMemory():
      self._memory_profiler = profilers.MemoryProfiler(
          self._name, configuration)
      self._memory_profiler.Start()

    if configuration.HaveProfileAnalyzers():
      self._analyzers_profiler = profilers.AnalyzersProfiler(
          f'{self._name:s}-analyzers', configuration)
      self._analyzers_profiler.Start()

    if configuration.HaveProfileProcessing():
      self._processing_profiler = profilers.ProcessingProfiler(
          f'{self._name:s}-processing', configuration)
      self._processing_profiler.Start()

    if configuration.HaveProfileSerializers():
      self._serializers_profiler = profilers.SerializersProfiler(
          f'{self._name:s}-serializers', configuration)
      self._serializers_profiler.Start()

    if configuration.HaveProfileStorage():
      self._storage_profiler = profilers.StorageProfiler(
          self._name, configuration)
      self._storage_profiler.Start()

    if configuration.HaveProfileTaskQueue():
      self._task_queue_profiler = profilers.TaskQueueProfiler(
          self._name, configuration)
      self._task_queue_profiler.Start()

  def _StopProfiling(self):
    """Stops profiling."""
    if self._memory_profiler:
      self._memory_profiler.Stop()
      self._memory_profiler = None

    if self._analyzers_profiler:
      self._analyzers_profiler.Stop()
      self._analyzers_profiler = None

    if self._processing_profiler:
      self._processing_profiler.Stop()
      self._processing_profiler = None

    if self._serializers_profiler:
      self._serializers_profiler.Stop()
      self._serializers_profiler = None

    if self._storage_profiler:
      self._storage_profiler.Stop()
      self._storage_profiler = None

    if self._task_queue_profiler:
      self._task_queue_profiler.Stop()
      self._task_queue_profiler = None

  def BuildArtifactsRegistry(
      self, artifact_definitions_path, custom_artifacts_path):
    """Builds an artificats definition registry.

    Args:
      artifact_definitions_path (str): path to artifact definitions directory
          or file.
      custom_artifacts_path (str): path to custom artifact definitions
          directory or file.

    Raises:
      BadConfigOption: if artifact definitions cannot be read.
    """
    if not artifact_definitions_path:
      raise errors.BadConfigOption('Missing artifact definitions path.')

    registry = artifacts_registry.ArtifactDefinitionsRegistry()
    reader = artifacts_reader.YamlArtifactsReader()

    try:
      if os.path.isdir(artifact_definitions_path):
        registry.ReadFromDirectory(reader, artifact_definitions_path)
      else:
        registry.ReadFromFile(reader, artifact_definitions_path)

    except (KeyError, artifacts_errors.FormatError) as exception:
      raise errors.BadConfigOption((
          f'Unable to read artifact definitions from: '
          f'{artifact_definitions_path:s} with error: {exception!s}'))

    if custom_artifacts_path:
      try:
        if os.path.isdir(custom_artifacts_path):
          registry.ReadFromDirectory(reader, custom_artifacts_path)
        else:
          registry.ReadFromFile(reader, custom_artifacts_path)

      except (KeyError, artifacts_errors.FormatError) as exception:
        raise errors.BadConfigOption((
            f'Unable to read custom artifact definitions from: '
            f'{custom_artifacts_path:s} with error: {exception!s}'))

    self._artifacts_registry = registry

  def BuildCollectionFilters(
      self, environment_variables, user_accounts,
      artifact_filter_names=None, filter_file_path=None,
      enable_artifacts_map=False):
    """Builds collection filters from artifacts or filter file if available.

    Args:
      environment_variables (list[EnvironmentVariableArtifact]):
          environment variables.
      user_accounts (list[UserAccountArtifact]): user accounts.
      artifact_filter_names (Optional[list[str]]): names of artifact
          definitions that are used for filtering file system and Windows
          Registry key paths.
      filter_file_path (Optional[str]): path of filter file.
      enable_artifacts_map (Optional[bool]): True if the artifacts path map
          should be generated. Defaults to False.

    Raises:
      InvalidFilter: if no valid file system find specifications are built.
    """
    filters_helper = None

    if artifact_filter_names:
      names = ', '.join(artifact_filter_names)
      logger.debug(f'building find specification based on artifacts: {names:s}')

      filters_helper = artifact_filters.ArtifactDefinitionsFiltersHelper(
          self._artifacts_registry)
      filters_helper.BuildFindSpecs(
          artifact_filter_names, environment_variables=environment_variables,
          user_accounts=user_accounts,
          enable_artifacts_map=enable_artifacts_map)

      # If the user selected Windows Registry artifacts we have to ensure
      # the Windows Registry files are parsed.
      if filters_helper.registry_find_specs:
        filters_helper.BuildFindSpecs(
            self._WINDOWS_REGISTRY_FILES_ARTIFACT_NAMES,
            environment_variables=environment_variables,
            user_accounts=user_accounts,
            enable_artifacts_map=enable_artifacts_map,
            original_registry_artifact_filter_names=(
                filters_helper.registry_find_specs_artifact_names))

      if not filters_helper.file_system_find_specs:
        raise errors.InvalidFilter(
            'No valid file system find specifications were built from '
            'artifacts.')

      self._included_file_system_find_specs = (
          filters_helper.file_system_find_specs)
      self._registry_find_specs = filters_helper.registry_find_specs

      self._artifacts_trie = filters_helper.artifacts_trie

    elif filter_file_path:
      logger.debug((
          f'building find specification based on filter file: '
          f'{filter_file_path:s}'))

      filter_file_object = yaml_filter_file.YAMLFilterFile()
      filter_file_path_filters = filter_file_object.ReadFromFile(
          filter_file_path)

      filters_helper = path_filters.PathCollectionFiltersHelper()
      filters_helper.BuildFindSpecs(
          filter_file_path_filters, environment_variables=environment_variables)

      if (not filters_helper.excluded_file_system_find_specs and
          not filters_helper.included_file_system_find_specs):
        raise errors.InvalidFilter((
            f'No valid file system find specifications were built from filter '
            f'file: {filter_file_path:s}.'))

      self._excluded_file_system_find_specs = (
          filters_helper.excluded_file_system_find_specs)
      self._included_file_system_find_specs = (
          filters_helper.included_file_system_find_specs)

  # pylint: disable=too-many-arguments
  @classmethod
  def CreateSession(
      cls, artifact_filter_names=None, command_line_arguments=None,
      debug_mode=False, filter_file_path=None, preferred_encoding='utf-8'):
    """Creates a session attribute container.

    Args:
      artifact_filter_names (Optional[list[str]]): names of artifact definitions
          that are used for filtering file system and Windows Registry
          key paths.
      command_line_arguments (Optional[str]): the command line arguments.
      debug_mode (Option[bool]): True if debug mode was enabled.
      filter_file_path (Optional[str]): path to a file with find specifications.
      preferred_encoding (Optional[str]): preferred encoding.

    Returns:
      Session: session attribute container.
    """
    session = sessions.Session()

    session.artifact_filters = artifact_filter_names
    session.command_line_arguments = command_line_arguments
    session.debug_mode = debug_mode
    session.filter_file = filter_file_path
    session.preferred_encoding = preferred_encoding

    return session

  def GetArtifactsTrie(self):
    """Retrieves the artifacts trie.

    Returns:
      ArtifactsTrie: artifacts trie.
    """
    return self._artifacts_trie

  def GetCollectionExcludedFindSpecs(self):
    """Retrieves find specifications to exclude from collection.

    Returns:
      list[dfvfs.FindSpec]: find specifications to exclude from collection.
    """
    return self._excluded_file_system_find_specs or []

  def GetCollectionIncludedFindSpecs(self):
    """Retrieves find specifications to include in collection.

    Returns:
      list[dfvfs.FindSpec]: find specifications to include in collection.
    """
    return self._included_file_system_find_specs or []

  def GetSourceFileSystem(self, file_system_path_spec, resolver_context=None):
    """Retrieves the file system of the source.

    Args:
      file_system_path_spec (dfvfs.PathSpec): path specifications of
          the source file system to process.
      resolver_context (dfvfs.Context): resolver context.

    Returns:
      tuple[dfvfs.FileSystem, path.PathSpec]: file system and mount point path
          specification. The mount point path specification refers to either a
          directory or a volume on a storage media device or image. It is
          needed by the dfVFS file system to indicate the base location of the
          file system.

    Raises:
      RuntimeError: if source file system path specification is not set.
    """
    if not file_system_path_spec:
      raise RuntimeError('Missing source file system path specification.')

    file_system = path_spec_resolver.Resolver.OpenFileSystem(
        file_system_path_spec, resolver_context=resolver_context)

    type_indicator = file_system_path_spec.type_indicator
    if path_spec_factory.Factory.IsSystemLevelTypeIndicator(type_indicator):
      mount_point = file_system_path_spec
    else:
      mount_point = file_system_path_spec.parent

    return file_system, mount_point

  def PreprocessSource(
      self, file_system_path_specs, storage_writer, resolver_context=None):
    """Preprocesses a source.

    Args:
      file_system_path_specs (list[dfvfs.PathSpec]): path specifications of
          the source file systems to process.
      storage_writer (StorageWriter): storage writer.
      resolver_context (Optional[dfvfs.Context]): resolver context.

    Returns:
      list[SystemConfigurationArtifact]: system configurations found in
          the source.
    """
    mediator = preprocess_mediator.PreprocessMediator(storage_writer)

    detected_operating_systems = []
    system_configurations = []
    for path_spec in file_system_path_specs:
      try:
        file_system, mount_point = self.GetSourceFileSystem(
            path_spec, resolver_context=resolver_context)
      except (RuntimeError, dfvfs_errors.BackEndError) as exception:
        logger.error(exception)
        continue

      preprocess_manager.PreprocessPluginsManager.RunPlugins(
          self._artifacts_registry, file_system, mount_point, mediator)

      operating_system = mediator.GetValue('operating_system')
      if not operating_system:
        continue

      detected_operating_systems.append(operating_system)

      system_configuration = artifacts.SystemConfigurationArtifact(
          code_page=mediator.code_page, language=mediator.language)
      # Ensure environment_variables is a list otherwise serialization will
      # fail.
      system_configuration.environment_variables = list(
          mediator.GetEnvironmentVariables())
      system_configuration.hostname = mediator.hostname
      system_configuration.keyboard_layout = mediator.GetValue(
          'keyboard_layout')
      system_configuration.operating_system = mediator.GetValue(
          'operating_system')
      system_configuration.operating_system_product = mediator.GetValue(
          'operating_system_product')
      system_configuration.operating_system_version = mediator.GetValue(
          'operating_system_version')
      # TODO: add support for multi file system system configurations.
      system_configuration.path_specs = [path_spec]

      if mediator.time_zone:
        system_configuration.time_zone = mediator.time_zone.zone

      system_configurations.append(system_configuration)

      mediator.Reset()

    if system_configurations:
      # TODO: kept for backwards compatibility.
      self.knowledge_base.ReadSystemConfigurationArtifact(
          system_configurations[0])
      for environment_variable in system_configuration.environment_variables:
        self.knowledge_base.AddEnvironmentVariable(environment_variable)

    return system_configurations

  def SetStatusUpdateInterval(self, status_update_interval):
    """Sets the status update interval.

    Args:
      status_update_interval (float): status update interval.
    """
    self._status_update_interval = status_update_interval



"""Manages custom event formatter helpers."""


class FormattersManager:
  """Custom event formatter helpers manager."""

  _custom_formatter_helpers = {}

  @classmethod
  def GetEventFormatterHelper(cls, identifier):
    """Retrieves a custom event formatter helper.

    Args:
      identifier (str): identifier.

    Returns:
      CustomEventFormatterHelper: custom event formatter or None if not
          available.
    """
    identifier = identifier.lower()
    return cls._custom_formatter_helpers.get(identifier)

  @classmethod
  def RegisterEventFormatterHelper(cls, formatter_helper_class):
    """Registers a custom event formatter helper.

    The custom event formatter helpers are identified based on their lower
    case identifier.

    Args:
      formatter_helper_class (type): class of the custom event formatter helper.

    Raises:
      KeyError: if a custom formatter helper is already set for the
          corresponding identifier.
    """
    identifier = formatter_helper_class.IDENTIFIER.lower()
    if identifier in cls._custom_formatter_helpers:
      raise KeyError((
          f'Custom event formatter helper already set for identifier: '
          f'{formatter_helper_class.IDENTIFIER:s}'))

    cls._custom_formatter_helpers[identifier] = formatter_helper_class()

  @classmethod
  def RegisterEventFormatterHelpers(cls, formatter_helper_classes):
    """Registers custom event formatter helpers.

    The formatter classes are identified based on their lower case data type.

    Args:
      formatter_helper_classes (list[type]): classes of the custom event
          formatter helpers.

    Raises:
      KeyError: if a custom formatter helper is already set for the
          corresponding data type.
    """
    for formatter_helper_class in formatter_helper_classes:
      cls.RegisterEventFormatterHelper(formatter_helper_class)



"""This file contains the interface for analysis plugins."""

import abc
import collections
import time

import requests

from plaso.analysis import interface
from plaso.analysis import logger
from plaso.containers import events
from plaso.lib import errors


class HashAnalysis:
  """Analysis information about a hash.

  Attributes:
    hash_information (object): object containing information about the hash.
    subject_hash (str): hash that was analyzed.
  """

  def __init__(self, subject_hash, hash_information):
    """Initializes analysis information about a hash.

    Args:
      subject_hash (str): hash that the hash_information relates to.
      hash_information (object): information about the hash. This object will be
          used by the GenerateLabels method in the HashTaggingAnalysisPlugin
          to tag events that relate to the hash.
    """
    self.hash_information = hash_information
    self.subject_hash = subject_hash


class HashTaggingAnalysisPlugin(interface.AnalysisPlugin):
  """An interface for plugins that tag events based on the source file hash."""

  # The event data types the plugin will collect hashes from. Subclasses
  # must override this attribute.
  DATA_TYPES = []

  # Lookup hashes supported by the hash tagging analysis plugin.
  SUPPORTED_HASHES = frozenset([])

  _DEFAULT_HASHES_PER_BATCH = 1
  _DEFAULT_LOOKUP_HASH = 'sha256'
  _DEFAULT_WAIT_AFTER_ANALYSIS = 0.0

  _REQUEST_TIMEOUT = 60

  def __init__(self):
    """Initializes a hash tagging analysis plugin."""
    super().__init__()
    self._batch_of_lookup_hashes = []
    self._data_stream_identifiers = set()
    self._data_streams_by_hash = collections.defaultdict(set)
    self._event_identifiers_by_data_stream = collections.defaultdict(set)
    self._hashes_per_batch = self._DEFAULT_HASHES_PER_BATCH
    self._lookup_hash = self._DEFAULT_LOOKUP_HASH
    self._wait_after_analysis = self._DEFAULT_WAIT_AFTER_ANALYSIS

  @abc.abstractmethod
  def _Analyze(self, hashes):
    """Analyzes a list of hashes.

    Args:
      hashes (list[str]): list of hashes to look up.

    Returns:
      list[HashAnalysis]: list of results of analyzing the hashes.
    """

  @abc.abstractmethod
  def _GenerateLabels(self, hash_information):
    """Generates a list of strings to tag events with.

    Args:
      hash_information (bool): response from the hash tagging analyzer that
          indicates that the file hash was present or not.

    Returns:
      list[str]: list of labels to apply to event.
    """

  def _MakeRequestAndDecodeJSON(self, url, method, **kwargs):
    """Make a HTTP request and decode the results as JSON.

    Args:
      url (str): URL to make a request to.
      method (str): HTTP method to used to make the request. GET and POST are
          supported.
      kwargs: parameters to the requests .get() or post() methods, depending
          on the value of the method parameter.

    Returns:
      dict[str, object]: body of the HTTP response, decoded from JSON.

    Raises:
      ConnectionError: If it is not possible to connect to the given URL, or it
          the request returns a HTTP error.
      ValueError: If an invalid HTTP method is specified.
    """
    method_upper = method.upper()
    if method_upper not in ('GET', 'POST'):
      raise ValueError('Method {0:s} is not supported')

    response = None

    try:
      if method_upper == 'GET':
        response = requests.get(url, timeout=self._REQUEST_TIMEOUT, **kwargs)

      elif method_upper == 'POST':
        response = requests.post(url, timeout=self._REQUEST_TIMEOUT, **kwargs)

      response.raise_for_status()

    except requests.ConnectionError as exception:
      raise errors.ConnectionError(
          f'Unable to connect to: {url:s} with error: {exception!s}')

    except requests.HTTPError as exception:
      raise errors.ConnectionError(
          f'Connect to: {url:s} returned a HTTP error: {exception!s}')

    return response.json()

  def _ProcessHashAnalysis(self, analysis_mediator, hash_analysis):
    """Processes the results of the analysis of a hash.

    This method ensures that labels are generated for the hash,
    then tags all events derived from files with that hash.

    Args:
      analysis_mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfVFS.
      hash_analysis (HashAnalysis): hash analysis plugin's results for a given
          hash.
    """
    labels = self._GenerateLabels(hash_analysis.hash_information)

    try:
      data_stream_identifiers = self._data_streams_by_hash.pop(
          hash_analysis.subject_hash)
    except KeyError:
      data_stream_identifiers = []
      logger.error((
          f'unable to retrieve data streams for digest hash: '
          f'{hash_analysis.subject_hash:s}'))

    for data_stream_identifier in data_stream_identifiers:
      event_identifiers = self._event_identifiers_by_data_stream.pop(
          data_stream_identifier)

      # Do no bail out earlier to maintain the state of
      # self._data_streams_by_hash and self._event_identifiers_by_data_stream.
      if not labels:
        continue

      for event_identifier in event_identifiers:
        event_tag = events.EventTag()
        event_tag.SetEventIdentifier(event_identifier)

        try:
          event_tag.AddLabels(labels)
        except (TypeError, ValueError):
          error_label = f'error_{self.NAME:s}'
          labels_string = ', '.join(labels)
          logger.error((
              f'unable to add labels: {labels_string!s} for digest hash: '
              f'{hash_analysis.subject_hash:s} defaulting to: {error_label:s}'))
          labels = [error_label]
          event_tag.AddLabels(labels)

        analysis_mediator.ProduceEventTag(event_tag)

        for label in labels:
          self._analysis_counter[label] += 1

  def CompileReport(self, analysis_mediator):
    """Compiles an analysis report.

    Args:
      analysis_mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfVFS.

    Returns:
      AnalysisReport: report.
    """
    if self._batch_of_lookup_hashes:
      for hash_analysis in self._Analyze(self._batch_of_lookup_hashes):
        self._ProcessHashAnalysis(analysis_mediator, hash_analysis)

      self._batch_of_lookup_hashes = []

    return super().CompileReport(
        analysis_mediator)

  def ExamineEvent(
      self, analysis_mediator, event, event_data, event_data_stream):
    """Evaluates whether an event contains the right data for a hash lookup.

    Args:
      analysis_mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
    """
    if (not self._lookup_hash or not event_data_stream or
        event_data.data_type not in self.DATA_TYPES):
      return

    data_stream_identifier = event_data_stream.GetIdentifier()
    if data_stream_identifier not in self._data_stream_identifiers:
      self._data_stream_identifiers.add(data_stream_identifier)

      lookup_hash = f'{self._lookup_hash:s}_hash'
      lookup_hash = getattr(event_data_stream, lookup_hash, None)
      if not lookup_hash:
        path_specification = getattr(event_data_stream, 'path_spec', None)
        display_name = analysis_mediator.GetDisplayNameForPathSpec(
            path_specification)
        logger.warning((
            f'Lookup hash attribute: {self._lookup_hash:s}_hash missing from '
            f'event data stream: {display_name:s}.'))

      else:
        self._data_streams_by_hash[lookup_hash].add(data_stream_identifier)
        self._batch_of_lookup_hashes.append(lookup_hash)

    event_identifier = event.GetIdentifier()
    self._event_identifiers_by_data_stream[data_stream_identifier].add(
        event_identifier)

    if len(self._batch_of_lookup_hashes) >= self._hashes_per_batch:
      for hash_analysis in self._Analyze(self._batch_of_lookup_hashes):
        self._ProcessHashAnalysis(analysis_mediator, hash_analysis)

      self._batch_of_lookup_hashes = []

      time.sleep(self._wait_after_analysis)

  def SetLookupHash(self, lookup_hash):
    """Sets the hash to query.

    Args:
      lookup_hash (str): name of the hash attribute to look up.

    Raises:
      ValueError: if the lookup hash is not supported.
    """
    if lookup_hash not in self.SUPPORTED_HASHES:
      raise ValueError(f'Unsupported lookup hash: {lookup_hash!s}')

    self._lookup_hash = lookup_hash



"""Task attribute container definition."""

import time
import uuid

from acstore.containers import interface
from acstore.containers import manager

from plaso.lib import definitions


class Task(interface.AttributeContainer):
  """Task attribute container.

  A task describes a piece of work for a multi processing worker process
  for example a task to process a path specification or to analyze an event.

  Attributes:
    aborted (bool): True if the task was aborted.
    completion_time (int): time that the task was completed. Contains the
        number of micro seconds since January 1, 1970, 00:00:00 UTC.
    file_entry_type (str): dfVFS type of the file entry the path specification
        is referencing.
    has_retry (bool): True if the task was previously abandoned and a retry
        task was created, False otherwise.
    identifier (str): unique identifier of the task.
    last_processing_time (int): the last time the task was marked as being
        processed as number of milliseconds since January 1, 1970, 00:00:00 UTC.
    merge_priority (int): priority used for the task storage file merge, where
        a lower value indicates a higher priority to merge.
    path_spec (dfvfs.PathSpec): path specification.
    session_identifier (str): the identifier of the session the task is part of.
    start_time (int): time that the task was started. Contains the number
        of micro seconds since January 1, 1970, 00:00:00 UTC.
    storage_file_size (int): size of the storage file in bytes.
    storage_format (str): the format the task results are to be stored in.
  """

  CONTAINER_TYPE = 'task'

  SCHEMA = {
      'aborted': 'bool',
      'completion_time': 'int',
      'file_entry_type': 'str',
      'has_retry': 'bool',
      'identifier': 'str',
      'last_processing_time': 'int',
      'merge_priority': 'int',
      'path_spec': 'dfvfs.PathSpec',
      'session_identifier': 'str',
      'start_time': 'int',
      'storage_file_size': 'int',
      'storage_format': 'str'}

  def __init__(self, session_identifier=None):
    """Initializes a task attribute container.

    Args:
      session_identifier (Optional[str]): identifier of the session the task
          is part of.
    """
    super().__init__()
    self.aborted = False
    self.completion_time = None
    self.file_entry_type = None
    self.has_retry = False
    self.identifier = f'{uuid.uuid4().hex:s}'
    self.last_processing_time = None
    self.merge_priority = None
    self.path_spec = None
    self.session_identifier = session_identifier
    self.start_time = int(time.time() * definitions.MICROSECONDS_PER_SECOND)
    self.storage_file_size = None
    self.storage_format = None

  # This method is necessary for heap sort.
  def __lt__(self, other):
    """Compares if the task attribute container is less than the other.

    Args:
      other (Task): task attribute container to compare to.

    Returns:
      bool: True if the task attribute container is less than the other.
    """
    return self.identifier < other.identifier

  def CreateRetryTask(self):
    """Creates a new task to retry a previously abandoned task.

    The retry task will have a new identifier but most of the attributes
    will be a copy of the previously abandoned task.

    Returns:
      Task: a task to retry a previously abandoned task.
    """
    retry_task = Task(session_identifier=self.session_identifier)
    retry_task.file_entry_type = self.file_entry_type
    retry_task.merge_priority = self.merge_priority
    retry_task.path_spec = self.path_spec
    retry_task.storage_file_size = self.storage_file_size
    retry_task.storage_format = self.storage_format

    self.has_retry = True

    return retry_task

  def UpdateProcessingTime(self):
    """Updates the processing time to now."""
    self.last_processing_time = int(
        time.time() * definitions.MICROSECONDS_PER_SECOND)


manager.AttributeContainersManager.RegisterAttributeContainer(Task)



"""JSON-L parser plugin for Microsoft (Office) 365 audit log files."""


from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import jsonl_parser
from plaso.parsers.jsonl_plugins import interface


class Microsoft365AuditLogEventData(events.EventData):
  """Microsoft (Office) 365 audit log event data.

  Attributes:
    audit_record_identifier (str): audit record identifier.
    application_access_context (str): application access context
    client_ip (str): client IP address.
    object_identifier (str): object identifier
    operation_name (str): operation name.
    organization_identifier (str): organization identifier.
    record_type (int): record type.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
    result_status (str): result status
    scope (str): scope.
    user_identifier (str): user identifier
    user_key (str): user key.
    user_type (int): user type.
    workload (str): Microsoft (Office) 365 service
  """

  DATA_TYPE = 'microsoft365:audit_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.audit_record_identifier = None
    self.application_access_context = None
    self.client_ip = None
    self.object_identifier = None
    self.operation_name = None
    self.organization_identifier = None
    self.record_type = None
    self.recorded_time = None
    self.result_status = None
    self.scope = None
    self.user_identifier = None
    self.user_key = None
    self.user_type = None
    self.workload = None


class Microsoft365AuditLogJSONLPlugin(interface.JSONLPlugin):
  """JSON-L parser plugin for Microsoft (Office) 365 audit log files."""

  NAME = 'microsoft_audit_log'
  DATA_FORMAT = 'Microsoft (Office) 365 audit log'

  def _ParseRecord(self, parser_mediator, json_dict):
    """Parses a Microsoft (Office) 365 audit log record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary of the log record.
    """
    date_time = None

    creation_time = self._GetJSONValue(json_dict, 'CreationTime')
    if creation_time:
      try:
        date_time = dfdatetime_time_elements.TimeElements()
        date_time.CopyFromStringISO8601(creation_time)
      except ValueError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'Unable to parse event time: {creation_time:s} with error: '
            f'{exception!s}')
        date_time = None

    event_data = Microsoft365AuditLogEventData()

    event_data.audit_record_identifier = self._GetJSONValue(json_dict, 'Id')
    event_data.application_access_context = self._GetJSONValue(
        json_dict, 'AppAccessContext')
    event_data.client_ip = self._GetJSONValue(json_dict, 'ClientIP')
    event_data.object_identifier = self._GetJSONValue(json_dict, 'ObjectId')
    event_data.operation_name = self._GetJSONValue(json_dict, 'Operation')
    event_data.organization_identifier = self._GetJSONValue(
        json_dict, 'OrganizationId')
    event_data.record_type = self._GetJSONValue(json_dict, 'RecordType')
    event_data.recorded_time = date_time
    event_data.result_status = self._GetJSONValue(json_dict, 'ResultStatus')
    event_data.scope = self._GetJSONValue(json_dict, 'Scope')
    event_data.user_identifier = self._GetJSONValue(json_dict, 'UserId')
    event_data.user_key = self._GetJSONValue(json_dict, 'UserKey')
    event_data.user_type = self._GetJSONValue(json_dict, 'UserType')
    event_data.workload = self._GetJSONValue(json_dict, 'Workload')

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, json_dict):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      json_dict (dict): JSON dictionary of the log record

    Returns:
      bool: True if this is the correct parsers, False otherwise.
    """
    audit_record_identifier = self._GetJSONValue(json_dict, 'Id')
    organization_identifier = self._GetJSONValue(json_dict, 'OrganizationId')
    creation_time = self._GetJSONValue(json_dict, 'CreationTime')

    if None in (audit_record_identifier, creation_time,
        organization_identifier):
      return False

    date_time = dfdatetime_time_elements.TimeElements()

    try:
      date_time.CopyFromStringISO8601(creation_time)
    except ValueError:
      return False

    return True


jsonl_parser.JSONLParser.RegisterPlugin(Microsoft365AuditLogJSONLPlugin)



"""Classes to assist in merging attribute containers of tasks."""

from plaso.containers import analysis_results
from plaso.containers import artifacts
from plaso.containers import event_sources
from plaso.containers import events
from plaso.containers import reports
from plaso.containers import warnings


class BaseTaskMergeHelper:
  """Interface of heler for merging task related attribute containers.

  Attributes:
    task_identifier (str): identifier of the task that is merged.
  """

  _CONTAINER_TYPES = ()

  def __init__(self, task_storage_reader, task_identifier):
    """Initialize a helper for merging task related attribute containers.

    Args:
      task_storage_reader (StorageReader): task storage reader.
      task_identifier (str): identifier of the task that is merged.
    """
    super().__init__()
    self._container_identifier_mappings = {}
    self._generator = self._GetAttributeContainers(task_storage_reader)
    self._task_storage_reader = task_storage_reader

    self.fully_merged = False
    self.task_identifier = task_identifier

  def _GetAttributeContainers(self, task_storage_reader):
    """Retrieves attribute containers to merge.

    Args:
      task_storage_reader (StorageReader): task storage reader.

    Yields:
      AttributeContainer: attribute container.
    """
    for container_type in self._CONTAINER_TYPES:
      yield from task_storage_reader.GetAttributeContainers(container_type)

    self.fully_merged = True

  def Close(self):
    """Closes the task storage reader."""
    self._task_storage_reader.Close()
    self._task_storage_reader = None

  def GetAttributeContainer(self):
    """Retrieves an attribute container to merge.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    try:
      container = next(self._generator)
    except StopIteration:
      container = None

    return container

  def GetAttributeContainerIdentifier(self, lookup_key):
    """Retrieves an attribute container.

    Args:
      lookup_key (str): lookup key that identifies the attribute container.

    Returns:
      AttributeContainerIdentifier: attribute container identifier that maps
          to the lookup key or None if not available.
    """
    return self._container_identifier_mappings.get(lookup_key)

  def SetAttributeContainerIdentifier(self, lookup_key, identifier):
    """Sets an attribute container.

    Args:
      lookup_key (str): lookup key that identifies the attribute container.
      identifier (AttributeContainerIdentifier): attribute container identifier.
    """
    self._container_identifier_mappings[lookup_key] = identifier


class AnalysisTaskMergeHelper(BaseTaskMergeHelper):
  """Assists in merging attribute containers of an analysis task."""

  # Container types produced by the analysis worker processes that need to be
  # merged. Note that some container types reference other container types and
  # therefore container types that are referenced, must be defined before
  # container types that reference them.

  _CONTAINER_TYPES = (
      events.EventTag.CONTAINER_TYPE,
      reports.AnalysisReport.CONTAINER_TYPE,
      warnings.AnalysisWarning.CONTAINER_TYPE,
      analysis_results.BrowserSearchAnalysisResult.CONTAINER_TYPE,
      analysis_results.ChromeExtensionAnalysisResult.CONTAINER_TYPE)


class ExtractionTaskMergeHelper(BaseTaskMergeHelper):
  """Assists in merging attribute containers of an extraction task."""

  # Container types produced by the extraction worker processes that need to be
  # merged. Note that some container types reference other container types and
  # therefore container types that are referenced, must be defined before
  # container types that reference them.

  _CONTAINER_TYPES = (
      event_sources.EventSource.CONTAINER_TYPE,
      events.EventDataStream.CONTAINER_TYPE,
      # The date-less log helper is needed to generate event from the event
      # data by the timeliner and therefore needs to be merged before event
      # data containers.
      events.DateLessLogHelper.CONTAINER_TYPE,
      events.EventData.CONTAINER_TYPE,
      warnings.ExtractionWarning.CONTAINER_TYPE,
      warnings.RecoveryWarning.CONTAINER_TYPE,
      artifacts.WindowsEventLogMessageFileArtifact.CONTAINER_TYPE,
      artifacts.WindowsEventLogMessageStringArtifact.CONTAINER_TYPE,
      artifacts.WindowsWevtTemplateEvent.CONTAINER_TYPE)



"""Windows Registry plugin to parse the RADAR Diagnosed Applications key."""

import os

from dfdatetime import filetime as dfdatetime_filetime
from dfdatetime import semantic_time as dfdatetime_semantic_time

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class DiagnosedApplicationsEventData(events.EventData):
  """RADAR Diagnosed Application event data attribute container.

  Attributes:
    key_path (str): Windows Registry key path.
    last_detection_time (dfdatetime.DateTimeValues): process last
        detected by RADAR date and time.
    last_written_time (dfdatetime.DateTimeValues): entry last written date
        and time.
    process_name (str): Name of the process diagnosed by RADAR.
  """

  DATA_TYPE = 'windows:registry:diagnosed_applications'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.key_path = None
    self.last_detection_time = None
    self.last_written_time = None
    self.process_name = None


class DiagnosedApplicationsPlugin(
    interface.WindowsRegistryPlugin, dtfabric_helper.DtFabricHelper):
  """RADAR Diagnosed Applications Windows Registry plugin."""

  NAME = 'diagnosed_applications'
  DATA_FORMAT = 'Diagnosed Applications Registry data'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\RADAR\\HeapLeakDetection\\'
          'DiagnosedApplications')])

  _DEFINITION_FILE = os.path.join(os.path.dirname(__file__), 'filetime.yaml')

  def _ParseFiletime(self, byte_stream):
    """Parses a FILETIME date and time value from a byte stream.

    Args:
      byte_stream (bytes): byte stream.

    Returns:
      dfdatetime.DateTimeValues: a FILETIME date and time values or a semantic
          date and time values if the FILETIME date and time value is not set.

    Raises:
      ParseError: if the FILETIME could not be parsed.
    """
    filetime_map = self._GetDataTypeMap('filetime')

    try:
      filetime = self._ReadStructureFromByteStream(
          byte_stream, 0, filetime_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError(
          f'Unable to parse FILETIME value with error: {exception!s}')

    if filetime == 0:
      return dfdatetime_semantic_time.NotSet()

    try:
      return dfdatetime_filetime.Filetime(timestamp=filetime)
    except ValueError:
      raise errors.ParseError(f'Invalid FILETIME value: 0x{filetime:08x}')

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    for subkey in registry_key.GetSubkeys():
      last_detection_value = subkey.GetValueByName('LastDetectionTime')
      date_time = self._ParseFiletime(last_detection_value.data)

      event_data = DiagnosedApplicationsEventData()
      event_data.key_path = subkey.path
      event_data.last_detection_time = date_time
      event_data.last_written_time = subkey.last_written_time
      event_data.process_name = subkey.name

      parser_mediator.ProduceEventData(event_data)


winreg_parser.WinRegistryParser.RegisterPlugin(DiagnosedApplicationsPlugin)
